@@ -7,25 +7,18 @@
 #include "kateexternaltoolsconfigwidget.h"
 #include "externaltoolsplugin.h"
 #include "kateexternaltool.h"
-#include "katetoolrunner.h"
 
 #include <KLineEdit>
-#include <KTextEditor/Document>
 #include <KTextEditor/Editor>
 #include <KTextEditor/View>
 #include <ktexteditor_version.h>
 
 #include <KConfig>
 #include <KConfigGroup>
-#include <KIconButton>
 #include <KIconLoader>
 #include <KMimeTypeChooser>
 #include <KSharedConfig>
-#include <KXMLGUIFactory>
-#include <KXmlGuiWindow>
 #include <QBitmap>
-#include <QComboBox>
-#include <QListView>
 #include <QMenu>
 #include <QMessageBox>
 #include <QRegularExpression>
