@@ -36,6 +36,12 @@ public:
     bool showFullPathOnRoots() const;
     void setShowFullPathOnRoots(bool);
 
+    bool showToolbar() const;
+    void setShowToolbar(bool);
+
+    bool showCloseButton() const;
+    void setShowCloseButton(bool);
+
 private:
     KConfigGroup m_group;
 
@@ -47,6 +53,8 @@ private:
     int m_sortRole;
 
     bool m_showFullPathOnRoots;
+    bool m_showToolbar;
+    bool m_showCloseButton;
 };
 
 #endif // KATE_FILETREE_PLUGIN_H
