@@ -14,6 +14,8 @@ namespace KTextEditor
 class Document;
 }
 
+enum { CustomSorting = 1000 };
+
 class KateFileTreeProxyModel : public QSortFilterProxyModel
 {
     Q_OBJECT
@@ -21,7 +23,9 @@ class KateFileTreeProxyModel : public QSortFilterProxyModel
 public:
     KateFileTreeProxyModel(QObject *p = nullptr);
     QModelIndex docIndex(const KTextEditor::Document *) const;
+    QModelIndex widgetIndex(QWidget *) const;
     bool isDir(const QModelIndex &i) const;
+    bool isWidgetDir(const QModelIndex &i) const;
     void setSourceModel(QAbstractItemModel *model) override;
     KTextEditor::Document *docFromIndex(const QModelIndex &index);
     QList<KTextEditor::Document *> docTreeFromIndex(const QModelIndex &index);
