@@ -24,10 +24,13 @@
 
 #include "plugin_katebuild.h"
 
+#include "AppOutput.h"
+
 #include "hostprocess.h"
 
 #include <cassert>
 
+#include <QApplication>
 #include <QCompleter>
 #include <QDir>
 #include <QFileDialog>
@@ -37,6 +40,7 @@
 #include <QRegularExpressionMatch>
 #include <QScrollBar>
 #include <QString>
+#include <QTimer>
 
 #include <QAction>
 
@@ -48,18 +52,23 @@
 #include <KTextEditor/MovingInterface>
 
 #include <KAboutData>
+#include <KColorScheme>
 #include <KLocalizedString>
 #include <KMessageBox>
 #include <KPluginFactory>
 #include <KXMLGUIFactory>
 
+#include <kterminallauncherjob.h>
 #include <ktexteditor_utils.h>
 
+#include <kde_terminal_interface.h>
+#include <kparts/part.h>
+
 K_PLUGIN_FACTORY_WITH_JSON(KateBuildPluginFactory, "katebuildplugin.json", registerPlugin<KateBuildPlugin>();)
 
 static const QString DefConfigCmd = QStringLiteral("cmake -DCMAKE_BUILD_TYPE=Debug -DCMAKE_INSTALL_PREFIX=/usr/local -DCMAKE_EXPORT_COMPILE_COMMANDS=1 ../");
 static const QString DefConfClean;
-static const QString DefTargetName = QStringLiteral("all");
+static const QString DefTargetName = QStringLiteral("build");
 static const QString DefBuildCmd = QStringLiteral("make");
 static const QString DefCleanCmd = QStringLiteral("make clean");
 static const QString NinjaPrefix = QStringLiteral("[ninja-detection]");
@@ -124,26 +133,29 @@ KateBuildView::KateBuildView(KTextEditor::Plugin *plugin, KTextEditor::MainWindo
     , m_filenameDetector(QStringLiteral("((?:[a-np-zA-Z]:[\\\\/])?[^\\s:(]+)[:\\(](\\d+)[,:]?(\\d+)?[\\):]* (.*)"))
     , m_newDirDetector(QStringLiteral("make\\[.+\\]: .+ '(.*)'"))
 {
-    KXMLGUIClient::setComponentName(QStringLiteral("katebuild"), i18n("Kate Build Plugin"));
+    KXMLGUIClient::setComponentName(QStringLiteral("katebuild"), i18n("Build"));
     setXMLFile(QStringLiteral("ui.rc"));
 
     m_toolView = mw->createToolView(plugin,
                                     QStringLiteral("kate_plugin_katebuildplugin"),
                                     KTextEditor::MainWindow::Bottom,
-                                    QIcon::fromTheme(QStringLiteral("application-x-ms-dos-executable")),
+                                    QIcon::fromTheme(QStringLiteral("run-build-clean")),
                                     i18n("Build"));
 
     QAction *a = actionCollection()->addAction(QStringLiteral("select_target"));
     a->setText(i18n("Select Target..."));
     a->setIcon(QIcon::fromTheme(QStringLiteral("select")));
     connect(a, &QAction::triggered, this, &KateBuildView::slotSelectTarget);
-    a = actionCollection()->addAction(QStringLiteral("build_default_target"));
-    a->setText(i18n("Build Default Target"));
-    connect(a, &QAction::triggered, this, &KateBuildView::slotBuildDefaultTarget);
 
-    a = actionCollection()->addAction(QStringLiteral("build_previous_target"));
-    a->setText(i18n("Build Previous Target"));
-    connect(a, &QAction::triggered, this, &KateBuildView::slotBuildPreviousTarget);
+    a = actionCollection()->addAction(QStringLiteral("build_selected_target"));
+    a->setText(i18n("Build Selected Target"));
+    a->setIcon(QIcon::fromTheme(QStringLiteral("run-build")));
+    connect(a, &QAction::triggered, this, &KateBuildView::slotBuildSelectedTarget);
+
+    a = actionCollection()->addAction(QStringLiteral("build_and_run_selected_target"));
+    a->setText(i18n("Build and Run Selected Target"));
+    a->setIcon(QIcon::fromTheme(QStringLiteral("media-playback-start")));
+    connect(a, &QAction::triggered, this, &KateBuildView::slotBuildAndRunSelectedTarget);
 
     a = actionCollection()->addAction(QStringLiteral("stop"));
     a->setText(i18n("Stop"));
@@ -169,9 +181,18 @@ KateBuildView::KateBuildView(KTextEditor::Plugin *plugin, KTextEditor::MainWindo
 
     m_buildWidget = new QWidget(m_toolView);
     m_buildUi.setupUi(m_buildWidget);
+    int leftMargin = QApplication::style()->pixelMetric(QStyle::PM_LayoutLeftMargin);
+    m_buildUi.u_outpTopLayout->setContentsMargins(leftMargin, 0, 0, 0);
     m_targetsUi = new TargetsUi(this, m_buildUi.u_tabWidget);
     m_buildUi.u_tabWidget->insertTab(0, m_targetsUi, i18nc("Tab label", "Target Settings"));
     m_buildUi.u_tabWidget->setCurrentWidget(m_targetsUi);
+    m_buildUi.u_tabWidget->setTabsClosable(true);
+    m_buildUi.u_tabWidget->tabBar()->tabButton(0, QTabBar::RightSide)->hide();
+    m_buildUi.u_tabWidget->tabBar()->tabButton(1, QTabBar::RightSide)->hide();
+    connect(m_buildUi.u_tabWidget, &QTabWidget::tabCloseRequested, this, [this](int index) {
+        // FIXME check if the process is still running
+        m_buildUi.u_tabWidget->widget(index)->deleteLater();
+    });
 
     m_buildWidget->installEventFilter(this);
 
@@ -219,8 +240,9 @@ KateBuildView::KateBuildView(KTextEditor::Plugin *plugin, KTextEditor::MainWindo
     connect(m_targetsUi->deleteTarget, &QToolButton::clicked, this, &KateBuildView::targetDelete);
 
     connect(m_targetsUi->addButton, &QToolButton::clicked, this, &KateBuildView::slotAddTargetClicked);
-    connect(m_targetsUi->buildButton, &QToolButton::clicked, this, &KateBuildView::slotBuildActiveTarget);
-    connect(m_targetsUi, &TargetsUi::enterPressed, this, &KateBuildView::slotBuildActiveTarget);
+    connect(m_targetsUi->buildButton, &QToolButton::clicked, this, &KateBuildView::slotBuildSelectedTarget);
+    connect(m_targetsUi->runButton, &QToolButton::clicked, this, &KateBuildView::slotBuildAndRunSelectedTarget);
+    connect(m_targetsUi, &TargetsUi::enterPressed, this, &KateBuildView::slotBuildAndRunSelectedTarget);
 
     m_proc.setOutputChannelMode(KProcess::SeparateChannels);
     connect(&m_proc, static_cast<void (QProcess::*)(int, QProcess::ExitStatus)>(&QProcess::finished), this, &KateBuildView::slotProcExited);
@@ -258,76 +280,78 @@ KateBuildView::~KateBuildView()
 void KateBuildView::readSessionConfig(const KConfigGroup &cg)
 {
     int numTargets = cg.readEntry(QStringLiteral("NumTargets"), 0);
+    m_projectTargetsetRow = cg.readEntry("ProjectTargetSetRow", 0);
     m_targetsUi->targetsModel.clear();
-    int tmpIndex;
-    int tmpCmd;
 
-    if (numTargets == 0) {
-        // either the config is empty or uses the older format
-        QModelIndex index = m_targetsUi->targetsModel.addTargetSet(i18n("Target Set"), QString());
-        m_targetsUi->targetsModel.addCommand(index, i18n("build"), cg.readEntry(QStringLiteral("Make Command"), DefBuildCmd));
-        m_targetsUi->targetsModel.addCommand(index, i18n("clean"), cg.readEntry(QStringLiteral("Clean Command"), DefCleanCmd));
-        m_targetsUi->targetsModel.addCommand(index, i18n("config"), DefConfigCmd);
+    for (int i = 0; i < numTargets; i++) {
+        QStringList targetNames = cg.readEntry(QStringLiteral("%1 Target Names").arg(i), QStringList());
+        QString targetSetName = cg.readEntry(QStringLiteral("%1 Target").arg(i), QString());
+        QString buildDir = cg.readEntry(QStringLiteral("%1 BuildPath").arg(i), QString());
 
-        QString quickCmd = cg.readEntry(QStringLiteral("Quick Compile Command"));
-        if (!quickCmd.isEmpty()) {
-            m_targetsUi->targetsModel.addCommand(index, i18n("quick"), quickCmd);
+        QModelIndex index = m_targetsUi->targetsModel.addTargetSet(targetSetName, buildDir);
+
+        // Keep a bit of backwards compatibility by ensuring that the "default" target is the first in the list
+        QString defCmd = cg.readEntry(QStringLiteral("%1 Target Default").arg(i), QString());
+        int defIndex = targetNames.indexOf(defCmd);
+        if (defIndex > 0) {
+            targetNames.move(defIndex, 0);
         }
-        tmpIndex = 0;
-        tmpCmd = 0;
-    } else {
-        for (int i = 0; i < numTargets; i++) {
-            QStringList targetNames = cg.readEntry(QStringLiteral("%1 Target Names").arg(i), QStringList());
-            QString targetSetName = cg.readEntry(QStringLiteral("%1 Target").arg(i), QString());
-            QString buildDir = cg.readEntry(QStringLiteral("%1 BuildPath").arg(i), QString());
-
-            QModelIndex index = m_targetsUi->targetsModel.addTargetSet(targetSetName, buildDir);
-
-            if (targetNames.isEmpty()) {
-                QString quickCmd = cg.readEntry(QStringLiteral("%1 QuickCmd").arg(i));
-                m_targetsUi->targetsModel.addCommand(index, i18n("build"), cg.readEntry(QStringLiteral("%1 BuildCmd"), DefBuildCmd));
-                m_targetsUi->targetsModel.addCommand(index, i18n("clean"), cg.readEntry(QStringLiteral("%1 CleanCmd"), DefCleanCmd));
-                if (!quickCmd.isEmpty()) {
-                    m_targetsUi->targetsModel.addCommand(index, i18n("quick"), quickCmd);
-                }
-                m_targetsUi->targetsModel.setDefaultCmd(i, i18n("build"));
-            } else {
-                for (int tn = 0; tn < targetNames.size(); ++tn) {
-                    const QString &targetName = targetNames.at(tn);
-                    m_targetsUi->targetsModel.addCommand(index, targetName, cg.readEntry(QStringLiteral("%1 BuildCmd %2").arg(i).arg(targetName), DefBuildCmd));
-                }
-                QString defCmd = cg.readEntry(QStringLiteral("%1 Target Default").arg(i), QString());
-                m_targetsUi->targetsModel.setDefaultCmd(i, defCmd);
-            }
+        for (int tn = 0; tn < targetNames.size(); ++tn) {
+            const QString &targetName = targetNames.at(tn);
+            const QString &buildCmd = cg.readEntry(QStringLiteral("%1 BuildCmd %2").arg(i).arg(targetName));
+            const QString &runCmd = cg.readEntry(QStringLiteral("%1 RunCmd %2").arg(i).arg(targetName));
+            m_targetsUi->targetsModel.addCommand(index, targetName, buildCmd, runCmd);
         }
-        tmpIndex = cg.readEntry(QStringLiteral("Active Target Index"), 0);
-        tmpCmd = cg.readEntry(QStringLiteral("Active Target Command"), 0);
     }
 
-    m_targetsUi->targetsView->expandAll();
-    m_targetsUi->targetsView->resizeColumnToContents(0);
-    m_targetsUi->targetsView->collapseAll();
-
-    QModelIndex root = m_targetsUi->targetsModel.index(tmpIndex);
-    QModelIndex cmdIndex = m_targetsUi->targetsModel.index(tmpCmd, 0, root);
-    cmdIndex = m_targetsUi->proxyModel.mapFromSource(cmdIndex);
-    m_targetsUi->targetsView->setCurrentIndex(cmdIndex);
-
     auto showMarks = cg.readEntry(QStringLiteral("Show Marks"), false);
     m_showMarks->setChecked(showMarks);
 
     // Add project targets, if any
-    slotAddProjectTarget();
+    addProjectTarget();
+
+    m_targetsUi->targetsView->expandAll();
+
+    // pre-select the last active target or the first target of the first set
+    int prevTargetSetRow = cg.readEntry(QStringLiteral("Active Target Index"), 0);
+    int prevCmdRow = cg.readEntry(QStringLiteral("Active Target Command"), 0);
+    QModelIndex rootIndex = m_targetsUi->targetsModel.index(prevTargetSetRow);
+    QModelIndex cmdIndex = m_targetsUi->targetsModel.index(prevCmdRow, 0, rootIndex);
+    cmdIndex = m_targetsUi->proxyModel.mapFromSource(cmdIndex);
+    m_targetsUi->targetsView->setCurrentIndex(cmdIndex);
+
+    m_targetsUi->updateTargetsButtonStates();
 }
 
 /******************************************************************/
 void KateBuildView::writeSessionConfig(KConfigGroup &cg)
 {
-    // Don't save project targets, is not our area of accountability
-    m_targetsUi->targetsModel.deleteTargetSet(i18n("Project Plugin Targets"));
+    // Save the active target
+    QModelIndex activeIndex = m_targetsUi->targetsView->currentIndex();
+    activeIndex = m_targetsUi->proxyModel.mapToSource(activeIndex);
+    if (activeIndex.isValid()) {
+        if (activeIndex.parent().isValid()) {
+            cg.writeEntry(QStringLiteral("Active Target Index"), activeIndex.parent().row());
+            cg.writeEntry(QStringLiteral("Active Target Command"), activeIndex.row());
+        } else {
+            cg.writeEntry(QStringLiteral("Active Target Index"), activeIndex.row());
+            cg.writeEntry(QStringLiteral("Active Target Command"), 0);
+        }
+    }
 
     QList<TargetModel::TargetSet> targets = m_targetsUi->targetsModel.targetSets();
 
+    // Don't save project target-set, but save the row index
+    m_projectTargetsetRow = 0;
+    for (int i = 0; i < targets.size(); i++) {
+        if (i18n("Project Plugin Targets") == targets[i].name) {
+            m_projectTargetsetRow = i;
+            targets.removeAt(i);
+            break;
+        }
+    }
+
+    cg.writeEntry("ProjectTargetSetRow", m_projectTargetsetRow);
     cg.writeEntry("NumTargets", targets.size());
 
     for (int i = 0; i < targets.size(); i++) {
@@ -336,34 +360,16 @@ void KateBuildView::writeSessionConfig(KConfigGroup &cg)
         QStringList cmdNames;
 
         for (int j = 0; j < targets[i].commands.count(); j++) {
-            const QString &cmdName = targets[i].commands[j].first;
-            const QString &buildCmd = targets[i].commands[j].second;
+            const QString &cmdName = targets[i].commands[j].name;
+            const QString &buildCmd = targets[i].commands[j].buildCmd;
+            const QString &runCmd = targets[i].commands[j].runCmd;
             cmdNames << cmdName;
             cg.writeEntry(QStringLiteral("%1 BuildCmd %2").arg(i).arg(cmdName), buildCmd);
+            cg.writeEntry(QStringLiteral("%1 RunCmd %2").arg(i).arg(cmdName), runCmd);
         }
         cg.writeEntry(QStringLiteral("%1 Target Names").arg(i), cmdNames);
-        cg.writeEntry(QStringLiteral("%1 Target Default").arg(i), targets[i].defaultCmd);
     }
-    int setRow = 0;
-    int set = 0;
-    QModelIndex ind = m_targetsUi->targetsView->currentIndex();
-    ind = m_targetsUi->proxyModel.mapToSource(ind);
-    if (ind.internalId() == TargetModel::InvalidIndex) {
-        set = ind.row();
-    } else {
-        set = ind.internalId();
-        setRow = ind.row();
-    }
-    if (setRow < 0) {
-        setRow = 0;
-    }
-
-    cg.writeEntry(QStringLiteral("Active Target Index"), set);
-    cg.writeEntry(QStringLiteral("Active Target Command"), setRow);
     cg.writeEntry(QStringLiteral("Show Marks"), m_showMarks->isChecked());
-
-    // Restore project targets, if any
-    slotAddProjectTarget();
 }
 
 /******************************************************************/
@@ -743,10 +749,10 @@ QUrl KateBuildView::docUrl()
 bool KateBuildView::checkLocal(const QUrl &dir)
 {
     if (dir.path().isEmpty()) {
-        KMessageBox::sorry(nullptr, i18n("There is no file or directory specified for building."));
+        KMessageBox::error(nullptr, i18n("There is no file or directory specified for building."));
         return false;
     } else if (!dir.isLocalFile()) {
-        KMessageBox::sorry(nullptr,
+        KMessageBox::error(nullptr,
                            i18n("The file \"%1\" is not a local file. "
                                 "Non-local files cannot be compiled.",
                                 dir.path()));
@@ -841,13 +847,51 @@ bool KateBuildView::slotStop()
 }
 
 /******************************************************************/
-void KateBuildView::slotBuildActiveTarget()
+void KateBuildView::slotBuildSelectedTarget()
 {
-    if (!m_targetsUi->targetsView->currentIndex().isValid()) {
+    QModelIndex currentIndex = m_targetsUi->targetsView->currentIndex();
+    if (!currentIndex.isValid() || (m_firstBuild && !m_targetsUi->targetsView->isVisible())) {
         slotSelectTarget();
-    } else {
-        buildCurrentTarget();
+        return;
     }
+    m_firstBuild = false;
+
+    if (!currentIndex.parent().isValid()) {
+        // This is a root item, try to build the first command
+        currentIndex = m_targetsUi->targetsView->model()->index(0, 0, currentIndex.siblingAtColumn(0));
+        if (currentIndex.isValid()) {
+            m_targetsUi->targetsView->setCurrentIndex(currentIndex);
+        } else {
+            slotSelectTarget();
+            return;
+        }
+    }
+    buildCurrentTarget();
+}
+
+/******************************************************************/
+void KateBuildView::slotBuildAndRunSelectedTarget()
+{
+    QModelIndex currentIndex = m_targetsUi->targetsView->currentIndex();
+    if (!currentIndex.isValid() || (m_firstBuild && !m_targetsUi->targetsView->isVisible())) {
+        slotSelectTarget();
+        return;
+    }
+    m_firstBuild = false;
+
+    if (!currentIndex.parent().isValid()) {
+        // This is a root item, try to build the first command
+        currentIndex = m_targetsUi->targetsView->model()->index(0, 0, currentIndex.siblingAtColumn(0));
+        if (currentIndex.isValid()) {
+            m_targetsUi->targetsView->setCurrentIndex(currentIndex);
+        } else {
+            slotSelectTarget();
+            return;
+        }
+    }
+
+    m_runAfterBuild = true;
+    buildCurrentTarget();
 }
 
 /******************************************************************/
@@ -861,59 +905,67 @@ void KateBuildView::slotBuildPreviousTarget()
     }
 }
 
-/******************************************************************/
-void KateBuildView::slotBuildDefaultTarget()
-{
-    QModelIndex defaultTarget = TargetModel::defaultTarget(m_targetsUi->targetsView->currentIndex());
-    m_targetsUi->targetsView->setCurrentIndex(defaultTarget);
-    buildCurrentTarget();
-}
-
 /******************************************************************/
 void KateBuildView::slotSelectTarget()
 {
     m_buildUi.u_tabWidget->setCurrentIndex(0);
     m_win->showToolView(m_toolView);
+    QPersistentModelIndex selected = m_targetsUi->targetsView->currentIndex();
     m_targetsUi->targetFilterEdit->setText(QString());
     m_targetsUi->targetFilterEdit->setFocus();
-    if (m_previousIndex.isValid()) {
-        m_targetsUi->targetsView->setCurrentIndex(m_previousIndex);
-    }
+
+    // Flash the target selection line-edit to show that something happened
+    // and where your focus went/should go
+    QPalette palette = m_targetsUi->targetFilterEdit->palette();
+    KColorScheme::adjustBackground(palette, KColorScheme::ActiveBackground);
+    m_targetsUi->targetFilterEdit->setPalette(palette);
+    QTimer::singleShot(500, this, [this]() {
+        m_targetsUi->targetFilterEdit->setPalette(QPalette());
+    });
+
     m_targetsUi->targetsView->expandAll();
+    if (selected.isValid()) {
+        m_targetsUi->targetsView->setCurrentIndex(selected);
+    }
+    m_targetsUi->targetsView->scrollTo(selected);
 }
 
 /******************************************************************/
 bool KateBuildView::buildCurrentTarget()
 {
-    if (m_proc.state() != QProcess::NotRunning) {
-        displayBuildResult(i18n("Already building..."), KTextEditor::Message::Warning);
-        return false;
-    }
-
     const QFileInfo docFInfo(docUrl().toLocalFile()); // docUrl() saves the current document
 
     QModelIndex ind = m_targetsUi->targetsView->currentIndex();
     m_previousIndex = ind;
     if (!ind.isValid()) {
-        KMessageBox::sorry(nullptr, i18n("No target available for building."));
+        KMessageBox::error(nullptr, i18n("No target available for building."));
         return false;
     }
 
     QString buildCmd = TargetModel::command(ind);
     QString cmdName = TargetModel::cmdName(ind);
-    m_searchPaths = TargetModel::workDir(ind).split(QLatin1Char(';'));
-    QString workDir = m_searchPaths.isEmpty() ? QString() : m_searchPaths.first();
+    m_searchPaths = TargetModel::searchPaths(ind);
+    QString workDir = TargetModel::workDir(ind);
     QString targetSet = TargetModel::targetName(ind);
 
     QString dir = workDir;
     if (workDir.isEmpty()) {
         dir = docFInfo.absolutePath();
         if (dir.isEmpty()) {
-            KMessageBox::sorry(nullptr, i18n("There is no local file or directory specified for building."));
+            KMessageBox::error(nullptr, i18n("There is no local file or directory specified for building."));
             return false;
         }
     }
 
+    if (m_proc.state() != QProcess::NotRunning) {
+        displayBuildResult(i18n("Already building..."), KTextEditor::Message::Warning);
+        return false;
+    }
+
+    if (m_runAfterBuild && buildCmd.isEmpty()) {
+        slotRunAfterBuild();
+        return true;
+    }
     // a single target can serve to build lots of projects with similar directory layout
     if (m_projectPluginView) {
         const QFileInfo baseDir(m_projectPluginView->property("projectBaseDir").toString());
@@ -996,10 +1048,10 @@ void KateBuildView::slotProcExited(int exitCode, QProcess::ExitStatus)
         m_buildUi.errTreeWidget->resizeColumnToContents(1);
         m_buildUi.errTreeWidget->resizeColumnToContents(2);
         m_buildUi.errTreeWidget->horizontalScrollBar()->setValue(0);
-        // m_buildUi.errTreeWidget->setSortingEnabled(true);
         m_win->showToolView(m_toolView);
     }
 
+    bool buildSuccess = true;
     if (m_numErrors || m_numWarnings) {
         QStringList msgs;
         if (m_numErrors) {
@@ -1011,6 +1063,8 @@ void KateBuildView::slotProcExited(int exitCode, QProcess::ExitStatus)
         }
         displayBuildResult(msgs.join(QLatin1Char('\n')), m_numErrors ? KTextEditor::Message::Error : KTextEditor::Message::Warning);
     } else if (exitCode != 0) {
+        buildSuccess = false;
+        m_runAfterBuild = false;
         displayBuildResult(i18n("Build failed."), KTextEditor::Message::Warning);
     } else {
         displayBuildResult(i18n("Build completed without problems."), KTextEditor::Message::Positive);
@@ -1023,6 +1077,101 @@ void KateBuildView::slotProcExited(int exitCode, QProcess::ExitStatus)
         // add marks
         slotViewChanged();
     }
+
+    if (buildSuccess && m_runAfterBuild) {
+        m_runAfterBuild = false;
+        slotRunAfterBuild();
+    }
+}
+
+void KateBuildView::slotRunAfterBuild()
+{
+    if (!m_previousIndex.isValid()) {
+        return;
+    }
+    QModelIndex idx = m_previousIndex;
+    QModelIndex runIdx = idx.siblingAtColumn(2);
+    const QString runCmd = runIdx.data().toString();
+    if (runCmd.isEmpty()) {
+        // Nothing to run, and not a problem
+        return;
+    }
+    const QString workDir = TargetModel::workDir(idx);
+    if (workDir.isEmpty()) {
+        displayBuildResult(i18n("Cannot execute: %1 No working directory set.", runCmd), KTextEditor::Message::Warning);
+        return;
+    }
+    QModelIndex nameIdx = idx.siblingAtColumn(0);
+    QString name = nameIdx.data().toString();
+
+    AppOutput *out = nullptr;
+    for (int i = 2; i < m_buildUi.u_tabWidget->count(); ++i) {
+        QString tabToolTip = m_buildUi.u_tabWidget->tabToolTip(i);
+        if (runCmd == tabToolTip) {
+            out = qobject_cast<AppOutput *>(m_buildUi.u_tabWidget->widget(i));
+            if (!out || !out->runningProcess().isEmpty()) {
+                out = nullptr;
+                continue;
+            }
+            // We have a winner, re-use this tab
+            m_buildUi.u_tabWidget->setCurrentIndex(i);
+            break;
+        }
+    }
+    if (!out) {
+        // This is means we create a new tab
+        out = new AppOutput();
+        int tabIndex = m_buildUi.u_tabWidget->addTab(out, name);
+        m_buildUi.u_tabWidget->setCurrentIndex(tabIndex);
+        m_buildUi.u_tabWidget->setTabToolTip(tabIndex, runCmd);
+        m_buildUi.u_tabWidget->setTabIcon(tabIndex, QIcon::fromTheme(QStringLiteral("media-playback-start")));
+
+        connect(out, &AppOutput::runningChanhged, this, [this]() {
+            // Update the tab icon when the run state changes
+            for (int i = 2; i < m_buildUi.u_tabWidget->count(); ++i) {
+                AppOutput *tabOut = qobject_cast<AppOutput *>(m_buildUi.u_tabWidget->widget(i));
+                if (!tabOut) {
+                    continue;
+                }
+                if (!tabOut->runningProcess().isEmpty()) {
+                    m_buildUi.u_tabWidget->setTabIcon(i, QIcon::fromTheme(QStringLiteral("media-playback-start")));
+                } else {
+                    m_buildUi.u_tabWidget->setTabIcon(i, QIcon::fromTheme(QStringLiteral("media-playback-pause")));
+                }
+            }
+        });
+    }
+
+    out->setWorkingDir(workDir);
+    out->runCommand(runCmd);
+
+    if (m_win->activeView()) {
+        m_win->activeView()->setFocus();
+    }
+}
+
+static void appendPlainTextTo(QPlainTextEdit *edit, const QString &text)
+{
+    // NOTE We can't use edit->appendPlainText() as that adds a new paragraph
+    // and we might need to add text that finishes the previous line.
+
+    // Get the scroll position to restore it if not at the end
+    int scrollValue = edit->verticalScrollBar()->value();
+    int scrollMax = edit->verticalScrollBar()->maximum();
+    // Save the selection and restore it after adding the text
+    QTextCursor cursor = edit->textCursor();
+
+    // Insert the new text
+    edit->moveCursor(QTextCursor::End);
+    edit->insertPlainText(text);
+
+    // Restore selection and scroll position
+    edit->setTextCursor(cursor);
+    if (scrollValue == scrollMax) {
+        edit->verticalScrollBar()->setValue(edit->verticalScrollBar()->maximum());
+    } else {
+        edit->verticalScrollBar()->setValue(scrollValue);
+    }
 }
 
 /******************************************************************/
@@ -1038,8 +1187,7 @@ void KateBuildView::slotReadReadyStdOut()
 
     // Remove the Ninja workaround string
     l.remove(NinjaPrefix);
-    m_buildUi.plainTextEdit->moveCursor(QTextCursor::End);
-    m_buildUi.plainTextEdit->insertPlainText(l);
+    appendPlainTextTo(m_buildUi.plainTextEdit, l);
 
     // handle one line at a time
     do {
@@ -1085,8 +1233,7 @@ void KateBuildView::slotReadReadyStdErr()
     // FIXME This works for utf8 but not for all charsets
     QString l = QString::fromUtf8(m_proc.readAllStandardError());
     l.remove(QLatin1Char('\r'));
-    m_buildUi.plainTextEdit->moveCursor(QTextCursor::End);
-    m_buildUi.plainTextEdit->insertPlainText(l);
+    appendPlainTextTo(m_buildUi.plainTextEdit, l);
     m_stdErr += l;
 
     do {
@@ -1154,12 +1301,16 @@ void KateBuildView::processLine(QStringView line)
 void KateBuildView::slotAddTargetClicked()
 {
     QModelIndex current = m_targetsUi->targetsView->currentIndex();
+    QString currName = DefTargetName;
+    QString currCmd;
+    QString currRun;
+
     if (current.parent().isValid()) {
+        // we need the root item
         current = current.parent();
     }
     current = m_targetsUi->proxyModel.mapToSource(current);
-
-    QModelIndex index = m_targetsUi->targetsModel.addCommand(current, DefTargetName, DefBuildCmd);
+    QModelIndex index = m_targetsUi->targetsModel.addCommand(current, currName, currCmd, currRun);
     index = m_targetsUi->proxyModel.mapFromSource(index);
     m_targetsUi->targetsView->setCurrentIndex(index);
 }
@@ -1169,10 +1320,10 @@ void KateBuildView::targetSetNew()
 {
     m_targetsUi->targetFilterEdit->setText(QString());
     QModelIndex index = m_targetsUi->targetsModel.addTargetSet(i18n("Target Set"), QString());
-    QModelIndex buildIndex = m_targetsUi->targetsModel.addCommand(index, i18n("Build"), DefBuildCmd);
-    m_targetsUi->targetsModel.addCommand(index, i18n("Clean"), DefCleanCmd);
-    m_targetsUi->targetsModel.addCommand(index, i18n("Config"), DefConfigCmd);
-    m_targetsUi->targetsModel.addCommand(index, i18n("ConfigClean"), DefConfClean);
+    QModelIndex buildIndex = m_targetsUi->targetsModel.addCommand(index, i18n("Build"), DefBuildCmd, QString());
+    m_targetsUi->targetsModel.addCommand(index, i18n("Clean"), DefCleanCmd, QString());
+    m_targetsUi->targetsModel.addCommand(index, i18n("Config"), DefConfigCmd, QString());
+    m_targetsUi->targetsModel.addCommand(index, i18n("ConfigClean"), DefConfClean, QString());
     buildIndex = m_targetsUi->proxyModel.mapFromSource(buildIndex);
     m_targetsUi->targetsView->setCurrentIndex(buildIndex);
 }
@@ -1260,7 +1411,7 @@ void KateBuildView::slotPluginViewCreated(const QString &name, QObject *pluginVi
     // add view
     if (pluginView && name == QLatin1String("kateprojectplugin")) {
         m_projectPluginView = pluginView;
-        slotAddProjectTarget();
+        addProjectTarget();
         connect(pluginView, SIGNAL(projectMapChanged()), this, SLOT(slotProjectMapChanged()), Qt::UniqueConnection);
     }
 }
@@ -1283,11 +1434,11 @@ void KateBuildView::slotProjectMapChanged()
         return;
     }
     m_targetsUi->targetsModel.deleteTargetSet(i18n("Project Plugin Targets"));
-    slotAddProjectTarget();
+    addProjectTarget();
 }
 
 /******************************************************************/
-void KateBuildView::slotAddProjectTarget()
+void KateBuildView::addProjectTarget()
 {
     // only do stuff with valid project
     if (!m_projectPluginView) {
@@ -1311,35 +1462,49 @@ void KateBuildView::slotAddProjectTarget()
     if (!projectsBaseDir.isEmpty()) {
         projectsBuildDir = QDir(projectsBaseDir).absoluteFilePath(projectsBuildDir);
     }
-    const QModelIndex set = m_targetsUi->targetsModel.addTargetSet(i18n("Project Plugin Targets"), projectsBuildDir);
+
+    m_projectTargetsetRow = std::min(m_projectTargetsetRow, m_targetsUi->targetsModel.rowCount());
+    const QModelIndex set = m_targetsUi->targetsModel.insertTargetSet(m_projectTargetsetRow, i18n("Project Plugin Targets"), projectsBuildDir);
+    const QString defaultTarget = buildMap.value(QStringLiteral("default_target")).toString();
 
     const QVariantList targetsets = buildMap.value(QStringLiteral("targets")).toList();
     for (const QVariant &targetVariant : targetsets) {
         QVariantMap targetMap = targetVariant.toMap();
         QString tgtName = targetMap[QStringLiteral("name")].toString();
         QString buildCmd = targetMap[QStringLiteral("build_cmd")].toString();
+        QString runCmd = targetMap[QStringLiteral("run_cmd")].toString();
 
         if (tgtName.isEmpty() || buildCmd.isEmpty()) {
             continue;
         }
-        m_targetsUi->targetsModel.addCommand(set, tgtName, buildCmd);
+        QPersistentModelIndex idx = m_targetsUi->targetsModel.addCommand(set, tgtName, buildCmd, runCmd);
+        if (tgtName == defaultTarget) {
+            // A bit of backwards compatibility, move the "default" target to the top
+            while (idx.row() > 0) {
+                m_targetsUi->targetsModel.moveRowUp(idx);
+            }
+        }
     }
 
-    if (!set.model()->index(0, 0, set).data().isValid()) {
+    if (!set.model()->index(0, 0, set).isValid()) {
         QString buildCmd = buildMap.value(QStringLiteral("build")).toString();
         QString cleanCmd = buildMap.value(QStringLiteral("clean")).toString();
         QString quickCmd = buildMap.value(QStringLiteral("quick")).toString();
         if (!buildCmd.isEmpty()) {
             // we have loaded an "old" project file (<= 4.12)
-            m_targetsUi->targetsModel.addCommand(set, i18n("build"), buildCmd);
+            m_targetsUi->targetsModel.addCommand(set, i18n("build"), buildCmd, QString());
         }
         if (!cleanCmd.isEmpty()) {
-            m_targetsUi->targetsModel.addCommand(set, i18n("clean"), cleanCmd);
+            m_targetsUi->targetsModel.addCommand(set, i18n("clean"), cleanCmd, QString());
         }
         if (!quickCmd.isEmpty()) {
-            m_targetsUi->targetsModel.addCommand(set, i18n("quick"), quickCmd);
+            m_targetsUi->targetsModel.addCommand(set, i18n("quick"), quickCmd, QString());
         }
     }
+    const auto index = m_targetsUi->proxyModel.mapFromSource(set);
+    if (index.isValid()) {
+        m_targetsUi->targetsView->expand(index);
+    }
 }
 
 /******************************************************************/
