@@ -16,7 +16,6 @@
 #include <KTextEditor/View>
 
 #include <QIcon>
-#include <QUrl>
 
 #include <algorithm>
 #include <utility>
@@ -104,13 +103,12 @@ class CompletionIcons : public QObject
 public:
     CompletionIcons()
         : QObject(KTextEditor::Editor::instance())
+        , classIcon(QIcon::fromTheme(QStringLiteral("code-class")))
+        , blockIcon(QIcon::fromTheme(QStringLiteral("code-block")))
+        , funcIcon(QIcon::fromTheme(QStringLiteral("code-function")))
+        , varIcon(QIcon::fromTheme(QStringLiteral("code-variable")))
+        , enumIcon(QIcon::fromTheme(QStringLiteral("enum")))
     {
-        classIcon = QIcon::fromTheme(QStringLiteral("code-class"));
-        blockIcon = QIcon::fromTheme(QStringLiteral("code-block"));
-        funcIcon = QIcon::fromTheme(QStringLiteral("code-function"));
-        varIcon = QIcon::fromTheme(QStringLiteral("code-variable"));
-        enumIcon = QIcon::fromTheme(QStringLiteral("enum"));
-
         auto e = KTextEditor::Editor::instance();
         QObject::connect(e, &KTextEditor::Editor::configChanged, this, [this](KTextEditor::Editor *e) {
             colorIcons(e);
@@ -196,6 +194,7 @@ class LSPClientCompletionImpl : public LSPClientCompletion
     QVector<QChar> m_triggersCompletion;
     QVector<QChar> m_triggersSignature;
     bool m_triggerSignature = false;
+    bool m_triggerCompletion = false;
 
     QList<LSPClientCompletionItem> m_matches;
     LSPClientServer::RequestHandle m_handle, m_handleSig;
@@ -299,6 +298,10 @@ public:
         qCInfo(LSPCLIENT) << "should start " << userInsertion << insertedText;
 
         if (!userInsertion || !m_server || insertedText.isEmpty()) {
+            if (!insertedText.isEmpty() && m_triggersSignature.contains(insertedText.back())) {
+                m_triggerSignature = true;
+                return true;
+            }
             return false;
         }
 
@@ -308,11 +311,11 @@ public:
 
         m_triggerSignature = false;
         complete = complete || m_triggersCompletion.contains(lastChar);
+        m_triggerCompletion = complete;
         if (m_triggersSignature.contains(lastChar)) {
             complete = true;
             m_triggerSignature = true;
         }
-
         return complete;
     }
 
@@ -322,11 +325,25 @@ public:
 
         qCInfo(LSPCLIENT) << "completion invoked" << m_server;
 
+        if (it == UserInvocation && range.isEmpty() && m_signatureHelp) {
+            // If this is a user invocation (ctrl-space), check the last non-space char for sig help trigger
+            QChar c;
+            int i = range.start().column() - 1;
+            int ln = range.start().line();
+            for (; i >= 0; --i) {
+                c = view->document()->characterAt(KTextEditor::Cursor(ln, i));
+                if (!c.isSpace()) {
+                    break;
+                }
+            }
+            m_triggerSignature = m_triggersSignature.contains(c);
+        }
+
         // maybe use WaitForReset ??
         // but more complex and already looks good anyway
-        auto handler = [this](const QList<LSPCompletionItem> & compl ) {
+        auto handler = [this](const QList<LSPCompletionItem> &completion) {
             beginResetModel();
-            qCInfo(LSPCLIENT) << "adding completions " << compl .size();
+            qCInfo(LSPCLIENT) << "adding completions " << completion.size();
             // purge all existing completion items
             m_matches.erase(std::remove_if(m_matches.begin(),
                                            m_matches.end(),
@@ -334,7 +351,7 @@ public:
                                                return ci.argumentHintDepth == 0;
                                            }),
                             m_matches.end());
-            for (const auto &item : compl ) {
+            for (const auto &item : completion) {
                 m_matches.push_back(item);
             }
             std::stable_sort(m_matches.begin(), m_matches.end(), compare_match);
@@ -378,10 +395,12 @@ public:
             auto position = view->cursorPosition();
             auto cursor = qMax(range.start(), qMin(range.end(), position));
             m_manager->update(document, false);
-            if (!m_triggerSignature) {
+
+            if (m_triggerCompletion) {
                 m_handle = m_server->documentCompletion(document->url(), {cursor.line(), cursor.column()}, this, handler);
             }
-            if (m_signatureHelp) {
+
+            if (m_signatureHelp && m_triggerSignature) {
                 m_handleSig = m_server->signatureHelp(document->url(), {cursor.line(), cursor.column()}, this, sigHandler);
             }
         }
