@@ -7,6 +7,7 @@
 #include "lspclientserver.h"
 
 #include "lspclient_debug.h"
+#include "hostprocess.h"
 
 #include <QCoreApplication>
 #include <QFileInfo>
@@ -32,6 +33,7 @@ static const QString MEMBER_VERSION = QStringLiteral("version");
 static const QString MEMBER_START = QStringLiteral("start");
 static const QString MEMBER_END = QStringLiteral("end");
 static const QString MEMBER_POSITION = QStringLiteral("position");
+static const QString MEMBER_POSITIONS = QStringLiteral("positions");
 static const QString MEMBER_LOCATION = QStringLiteral("location");
 static const QString MEMBER_RANGE = QStringLiteral("range");
 static const QString MEMBER_LINE = QStringLiteral("line");
@@ -117,6 +119,15 @@ static QJsonArray to_json(const QList<LSPTextDocumentContentChangeEvent> &change
     return result;
 }
 
+static QJsonArray to_json(const QVector<LSPPosition> &positions)
+{
+    QJsonArray result;
+    for (const auto &position : positions) {
+        result.push_back(to_json(position));
+    }
+    return result;
+}
+
 static QJsonObject versionedTextDocumentIdentifier(const QUrl &document, int version = -1)
 {
     QJsonObject map{{MEMBER_URI, document.toString()}};
@@ -151,6 +162,13 @@ static QJsonObject textDocumentPositionParams(const QUrl &document, LSPPosition
     return params;
 }
 
+static QJsonObject textDocumentPositionsParams(const QUrl &document, const QVector<LSPPosition> &positions)
+{
+    auto params = textDocumentParams(document);
+    params[MEMBER_POSITIONS] = to_json(positions);
+    return params;
+}
+
 static QJsonObject referenceParams(const QUrl &document, LSPPosition pos, bool decl)
 {
     auto params = textDocumentPositionParams(document, pos);
@@ -372,6 +390,7 @@ static void from_json(LSPServerCapabilities &caps, const QJsonObject &json)
     from_json(caps.semanticTokenProvider, json.value(QStringLiteral("semanticTokensProvider")).toObject());
     auto workspace = json.value(QStringLiteral("workspace")).toObject();
     from_json(caps.workspaceFolders, workspace.value(QStringLiteral("workspaceFolders")));
+    caps.selectionRangeProvider = toBoolOrObject(json.value(QStringLiteral("selectionRangeProvider")));
 }
 
 // follow suit; as performed in kate docmanager
@@ -391,6 +410,15 @@ static QUrl normalizeUrl(const QUrl &url)
     return url.adjusted(QUrl::NormalizePathSegments);
 }
 
+static void from_json(LSPVersionedTextDocumentIdentifier &id, const QJsonValue &json)
+{
+    if (json.isObject()) {
+        auto ob = json.toObject();
+        id.uri = normalizeUrl(QUrl(ob.value(MEMBER_URI).toString()));
+        id.version = ob.value(MEMBER_VERSION).toInt(-1);
+    }
+}
+
 static LSPResponseError parseResponseError(const QJsonValue &v)
 {
     LSPResponseError ret;
@@ -441,6 +469,37 @@ static LSPRange parseRange(const QJsonObject &range)
     return {startpos, endpos};
 }
 
+static std::shared_ptr<LSPSelectionRange> parseSelectionRange(QJsonValueRef selectionRange)
+{
+    auto current = std::make_shared<LSPSelectionRange>(LSPSelectionRange{});
+    std::shared_ptr<LSPSelectionRange> ret = current;
+    QJsonValue selRange = std::move(selectionRange);
+
+    while (selRange.isObject()) {
+        current->range = parseRange(selRange[MEMBER_RANGE].toObject());
+        if (!selRange[QStringLiteral("parent")].isObject()) {
+            current->parent = nullptr;
+            break;
+        }
+        selRange = selRange[QStringLiteral("parent")].toObject();
+        current->parent = std::make_shared<LSPSelectionRange>(LSPSelectionRange{});
+        current = current->parent;
+    }
+
+    return ret;
+}
+
+static QList<std::shared_ptr<LSPSelectionRange>> parseSelectionRanges(const QJsonValue &result)
+{
+    QList<std::shared_ptr<LSPSelectionRange>> ret;
+    auto selectionRanges = result.toArray();
+    for (QJsonValueRef selectionRange : selectionRanges) {
+        ret.push_back(parseSelectionRange(selectionRange));
+    }
+
+    return ret;
+}
+
 static LSPLocation parseLocation(const QJsonObject &loc)
 {
     auto uri = normalizeUrl(QUrl(loc.value(MEMBER_URI).toString()));
@@ -461,6 +520,19 @@ static LSPLocation parseLocationLink(const QJsonObject &loc)
     return {QUrl(uri), range};
 }
 
+static QList<LSPTextEdit> parseTextEdit(const QJsonValue &result)
+{
+    QList<LSPTextEdit> ret;
+    const auto textEdits = result.toArray();
+    for (const auto &redit : textEdits) {
+        auto edit = redit.toObject();
+        auto text = edit.value(QStringLiteral("newText")).toString();
+        auto range = parseRange(edit.value(MEMBER_RANGE).toObject());
+        ret.push_back({range, text});
+    }
+    return ret;
+}
+
 static LSPDocumentHighlight parseDocumentHighlight(const QJsonValue &result)
 {
     auto hover = result.toObject();
@@ -488,24 +560,7 @@ static QList<LSPDocumentHighlight> parseDocumentHighlightList(const QJsonValue &
 
 static LSPMarkupContent parseHoverContentElement(const QJsonValue &contents)
 {
-    LSPMarkupContent result;
-    if (contents.isString()) {
-        result.value = contents.toString();
-    } else {
-        // should be object, pretend so
-        auto cont = contents.toObject();
-        auto text = cont.value(QStringLiteral("value")).toString();
-        if (text.isEmpty()) {
-            // nothing to lose, try markdown
-            result = parseMarkupContent(contents);
-        } else {
-            result.value = text;
-        }
-    }
-    if (result.value.length()) {
-        result.kind = LSPMarkupKind::PlainText;
-    }
-    return result;
+    return parseMarkupContent(contents);
 }
 
 static LSPHover parseHover(const QJsonValue &result)
@@ -644,7 +699,10 @@ static QList<LSPCompletionItem> parseDocumentCompletion(const QJsonValue &result
         }
         auto kind = static_cast<LSPCompletionItemKind>(item.value(MEMBER_KIND).toInt());
         //         auto textEdit = parseTextEdit(item.value(QStringLiteral("textEdit")).toObject());
-        ret.push_back({label, kind, detail, doc, sortText, insertText /*, textEdit*/});
+
+        const auto additionalTextEdits = parseTextEdit(item.value(QStringLiteral("additionalTextEdits")));
+
+        ret.push_back({label, kind, detail, doc, sortText, insertText, additionalTextEdits /*, textEdit*/});
     }
     return ret;
 }
@@ -709,27 +767,37 @@ static QString parseClangdSwitchSourceHeader(const QJsonValue &result)
     return result.toString();
 }
 
-static QList<LSPTextEdit> parseTextEdit(const QJsonValue &result)
+static LSPExpandedMacro parseExpandedMacro(const QJsonValue &result)
 {
-    QList<LSPTextEdit> ret;
-    const auto textEdits = result.toArray();
-    for (const auto &redit : textEdits) {
-        auto edit = redit.toObject();
-        auto text = edit.value(QStringLiteral("newText")).toString();
-        auto range = parseRange(edit.value(MEMBER_RANGE).toObject());
-        ret.push_back({range, text});
-    }
+    LSPExpandedMacro ret;
+    const QJsonObject ob = result.toObject();
+    ret.name = ob.value(QStringLiteral("name")).toString();
+    ret.expansion = ob.value(QStringLiteral("expansion")).toString();
+    return ret;
+}
+
+static LSPTextDocumentEdit parseTextDocumentEdit(const QJsonValue &result)
+{
+    LSPTextDocumentEdit ret;
+    auto ob = result.toObject();
+    from_json(ret.textDocument, ob.value(QStringLiteral("textDocument")));
+    ret.edits = parseTextEdit(ob.value(QStringLiteral("edits")));
     return ret;
 }
 
 static LSPWorkspaceEdit parseWorkSpaceEdit(const QJsonValue &result)
 {
-    QHash<QUrl, QList<LSPTextEdit>> ret;
+    LSPWorkspaceEdit ret;
     auto changes = result.toObject().value(QStringLiteral("changes")).toObject();
     for (auto it = changes.begin(); it != changes.end(); ++it) {
-        ret.insert(normalizeUrl(QUrl(it.key())), parseTextEdit(it.value()));
+        ret.changes.insert(normalizeUrl(QUrl(it.key())), parseTextEdit(it.value()));
     }
-    return {ret};
+    auto documentChanges = result.toObject().value(QStringLiteral("documentChanges")).toArray();
+    // resourceOperations not supported for now
+    for (auto edit : documentChanges) {
+        ret.documentChanges.push_back(parseTextDocumentEdit(edit));
+    }
+    return ret;
 }
 
 static LSPCommand parseCommand(const QJsonObject &result)
@@ -966,7 +1034,10 @@ class LSPClientServer::LSPClientServerPrivate
     QHash<int, std::pair<GenericReplyHandler, GenericReplyHandler>> m_handlers;
     // pending request responses
     static constexpr int MAX_REQUESTS = 5;
-    QVector<int> m_requests{MAX_REQUESTS + 1};
+    QVector<QJsonValue> m_requests{MAX_REQUESTS + 1};
+
+    // currently accumulated stderr output, used to output to the message view on line level
+    QString m_currentStderrOutput;
 
 public:
     LSPClientServerPrivate(LSPClientServer *_q,
@@ -983,7 +1054,8 @@ public:
         , m_folders(folders)
     {
         // setup async reading
-        QObject::connect(&m_sproc, &QProcess::readyRead, utils::mem_fun(&self_type::read, this));
+        QObject::connect(&m_sproc, &QProcess::readyReadStandardOutput, utils::mem_fun(&self_type::readStandardOutput, this));
+        QObject::connect(&m_sproc, &QProcess::readyReadStandardError, utils::mem_fun(&self_type::readStandardError, this));
         QObject::connect(&m_sproc, &QProcess::stateChanged, utils::mem_fun(&self_type::onStateChanged, this));
     }
 
@@ -1035,7 +1107,7 @@ private:
         }
     }
 
-    RequestHandle write(const QJsonObject &msg, const GenericReplyHandler &h = nullptr, const GenericReplyHandler &eh = nullptr, const int *id = nullptr)
+    RequestHandle write(const QJsonObject &msg, const GenericReplyHandler &h = nullptr, const GenericReplyHandler &eh = nullptr, const QJsonValue *id = nullptr)
     {
         RequestHandle ret;
         ret.m_server = q;
@@ -1080,7 +1152,7 @@ private:
         return RequestHandle();
     }
 
-    void read()
+    void readStandardOutput()
     {
         // accumulate in buffer
         m_receive.append(m_sproc.readAllStandardOutput());
@@ -1183,6 +1255,26 @@ private:
         }
     }
 
+    void readStandardError()
+    {
+        // append new stuff to our buffer, we assume UTF-8 output
+        m_currentStderrOutput.append(QString::fromUtf8(m_sproc.readAllStandardError()));
+
+        // now, cut out all full lines
+        LSPShowMessageParams msg;
+        if (const int lastNewLineIndex = m_currentStderrOutput.lastIndexOf(QLatin1Char('\n')); lastNewLineIndex >= 0) {
+            msg.message = m_currentStderrOutput.left(lastNewLineIndex);
+            m_currentStderrOutput.remove(0, lastNewLineIndex + 1);
+        }
+
+        // emit the output lines if non-empty
+        // this might strip empty lines in error output, but that is better then a spammed output view
+        if (!msg.message.isEmpty()) {
+            msg.type = LSPMessageType::Log;
+            Q_EMIT q->logMessage(msg);
+        }
+    }
+
     static QJsonObject init_error(const LSPErrorCode code, const QString &msg)
     {
         return QJsonObject{{MEMBER_ERROR, QJsonObject{{MEMBER_CODE, static_cast<int>(code)}, {MEMBER_MESSAGE, msg}}}};
@@ -1257,6 +1349,13 @@ private:
                                             {QStringLiteral("codeAction"), codeAction},
                                             {QStringLiteral("semanticTokens"), semanticTokens},
                                             {QStringLiteral("synchronization"), QJsonObject{{QStringLiteral("didSave"), true}}},
+                                            {QStringLiteral("selectionRange"), QJsonObject{{QStringLiteral("dynamicRegistration"), false}}},
+                                            {QStringLiteral("hover"), QJsonObject{
+                                                {QStringLiteral("contentFormat"), QJsonArray{
+                                                    QStringLiteral("markdown"),
+                                                    QStringLiteral("plaintext")
+                                                }}
+                                            }}
                                         },
                                   },
                                   {QStringLiteral("window"),
@@ -1292,7 +1391,7 @@ private:
     }
 
 public:
-    bool start()
+    bool start(bool forwardStdError)
     {
         if (m_state != State::None) {
             return true;
@@ -1306,10 +1405,10 @@ public:
         // start LSP server in project root
         m_sproc.setWorkingDirectory(m_root.toLocalFile());
 
-        // at least we see some errors somewhere then
-        m_sproc.setProcessChannelMode(QProcess::ForwardedErrorChannel);
+        // we handle stdout/stderr internally, important stuff via stdout
+        m_sproc.setProcessChannelMode(forwardStdError ? QProcess::ForwardedErrorChannel : QProcess::SeparateChannels);
         m_sproc.setReadChannel(QProcess::QProcess::StandardOutput);
-        m_sproc.start(program, args);
+        startHostProcess(m_sproc, program, args);
         const bool result = m_sproc.waitForStarted();
         if (result) {
             setState(State::Started);
@@ -1392,6 +1491,12 @@ public:
         return send(init_request(QStringLiteral("textDocument/signatureHelp"), params), h);
     }
 
+    RequestHandle selectionRange(const QUrl &document, const QVector<LSPPosition> &positions, const GenericReplyHandler &h)
+    {
+        auto params = textDocumentPositionsParams(document, positions);
+        return send(init_request(QStringLiteral("textDocument/selectionRange"), params), h);
+    }
+
     RequestHandle clangdSwitchSourceHeader(const QUrl &document, const GenericReplyHandler &h)
     {
         auto params = QJsonObject{{MEMBER_URI, document.toString()}};
@@ -1403,6 +1508,12 @@ public:
         return send(init_request(QStringLiteral("$/memoryUsage"), QJsonObject()), h);
     }
 
+    RequestHandle rustAnalyzerExpandMacro(const QUrl &document, const LSPPosition &pos, const GenericReplyHandler &h)
+    {
+        auto params = textDocumentPositionParams(document, pos);
+        return send(init_request(QStringLiteral("rust-analyzer/expandMacro"), params), h);
+    }
+
     RequestHandle documentFormatting(const QUrl &document, const LSPFormattingOptions &options, const GenericReplyHandler &h)
     {
         auto params = documentRangeFormattingParams(document, nullptr, options);
@@ -1522,7 +1633,7 @@ public:
         }
     }
 
-    GenericReplyHandler prepareResponse(int msgid)
+    GenericReplyHandler prepareResponse(QJsonValue msgid)
     {
         // allow limited number of outstanding requests
         auto ctx = QPointer<LSPClientServer>(q);
@@ -1558,7 +1669,8 @@ public:
     void processRequest(const QJsonObject &msg)
     {
         auto method = msg[MEMBER_METHOD].toString();
-        auto msgid = msg[MEMBER_ID].toInt();
+        // could be number or string, let's retain as-is
+        auto msgid = msg[MEMBER_ID];
         auto params = msg[MEMBER_PARAMS];
         bool handled = false;
         if (method == QLatin1String("workspace/applyEdit")) {
@@ -1640,9 +1752,9 @@ const LSPServerCapabilities &LSPClientServer::capabilities() const
     return d->capabilities();
 }
 
-bool LSPClientServer::start()
+bool LSPClientServer::start(bool forwardStdError)
 {
-    return d->start();
+    return d->start(forwardStdError);
 }
 
 void LSPClientServer::stop(int to_t, int to_k)
@@ -1715,6 +1827,12 @@ LSPClientServer::signatureHelp(const QUrl &document, const LSPPosition &pos, con
     return d->signatureHelp(document, pos, make_handler(h, context, parseSignatureHelp));
 }
 
+LSPClientServer::RequestHandle
+LSPClientServer::selectionRange(const QUrl &document, const QVector<LSPPosition> &positions, const QObject *context, const SelectionRangeReplyHandler &h)
+{
+    return d->selectionRange(document, positions, make_handler(h, context, parseSelectionRanges));
+}
+
 LSPClientServer::RequestHandle LSPClientServer::clangdSwitchSourceHeader(const QUrl &document, const QObject *context, const SwitchSourceHeaderHandler &h)
 {
     return d->clangdSwitchSourceHeader(document, make_handler(h, context, parseClangdSwitchSourceHeader));
@@ -1728,6 +1846,12 @@ LSPClientServer::RequestHandle LSPClientServer::clangdMemoryUsage(const QObject
     return d->clangdMemoryUsage(make_handler(h, context, identity));
 }
 
+LSPClientServer::RequestHandle
+LSPClientServer::rustAnalyzerExpandMacro(const QObject *context, const QUrl &document, const LSPPosition &pos, const ExpandMacroHandler &h)
+{
+    return d->rustAnalyzerExpandMacro(document, pos, make_handler(h, context, parseExpandedMacro));
+}
+
 LSPClientServer::RequestHandle
 LSPClientServer::documentFormatting(const QUrl &document, const LSPFormattingOptions &options, const QObject *context, const FormattingReplyHandler &h)
 {
