@@ -6,10 +6,8 @@
 #include "lsptooltip.h"
 
 #include <QApplication>
-#include <QDebug>
 #include <QEvent>
 #include <QFontMetrics>
-#include <QLabel>
 #include <QMouseEvent>
 #include <QPointer>
 #include <QScreen>
@@ -21,7 +19,6 @@
 #include <KSyntaxHighlighting/Definition>
 #include <KSyntaxHighlighting/Repository>
 #include <KSyntaxHighlighting/SyntaxHighlighter>
-#include <KTextEditor/ConfigInterface>
 #include <KTextEditor/Editor>
 #include <KTextEditor/View>
 #include <KWindowSystem>
@@ -77,7 +74,7 @@ public:
         setWindowFlags(Qt::FramelessWindowHint | Qt::BypassGraphicsProxyWidget | Qt::ToolTip);
         setAttribute(Qt::WA_DeleteOnClose, true);
         document()->setDocumentMargin(5);
-        setFrameStyle(QFrame::Box | QFrame::Raised);
+        setFrameStyle(QFrame::NoFrame);
         connect(&m_hideTimer, &QTimer::timeout, this, &Tooltip::hideTooltip);
 
         setHorizontalScrollBarPolicy(Qt::ScrollBarAsNeeded);
