@@ -10,6 +10,7 @@
 // BEGIN Includes
 #include "katemainwindow.h"
 
+#include "filehistorywidget.h"
 #include "kateapp.h"
 #include "kateconfigdialog.h"
 #include "kateconfigplugindialogpage.h"
@@ -26,8 +27,8 @@
 #include "katestashmanager.h"
 #include "kateupdatedisabler.h"
 #include "kateviewspace.h"
+#include "ktexteditor_utils.h"
 
-#include <KAboutData>
 #include <KActionCollection>
 #include <KActionMenu>
 #include <KApplicationTrader>
@@ -38,7 +39,13 @@
 #include <KHelpClient>
 #include <KIO/ApplicationLauncherJob>
 #include <KIO/Job>
+#include <kio_version.h>
+#include <kwidgetsaddons_version.h>
+#if KIO_VERSION >= QT_VERSION_CHECK(5, 98, 0)
+#include <KIO/JobUiDelegateFactory>
+#else
 #include <KIO/JobUiDelegate>
+#endif
 #include <KLocalizedString>
 #include <KMessageBox>
 #include <KMultiTabBar>
@@ -57,20 +64,31 @@
 #include <QApplication>
 #include <QDir>
 #include <QFontDatabase>
+#include <QKeySequence>
 #include <QList>
 #include <QMenu>
 #include <QMenuBar>
 #include <QMimeData>
 #include <QMimeDatabase>
 #include <QScreen>
-#include <QStyle>
+#include <QStackedWidget>
 #include <QTimer>
 #include <QToolButton>
 
+#include "diffwidget.h"
+
 #include <ktexteditor/sessionconfiginterface.h>
 
 // END
 
+// shall windows close the documents only visible inside them if the are closed?
+static bool winClosesDocuments()
+{
+    const auto config = KSharedConfig::openConfig();
+    const KConfigGroup cgGeneral(config, "General");
+    return cgGeneral.readEntry("Close documents with window", true);
+}
+
 KateMwModOnHdDialog *KateMainWindow::s_modOnHdDialog = nullptr;
 
 KateContainerStackedLayout::KateContainerStackedLayout(QWidget *parent)
@@ -160,14 +178,13 @@ KateMainWindow::KateMainWindow(KConfig *sconfig, const QString &sgroup)
     // Hence, we have to take care of the menu bar here
     toggleShowMenuBar(false);
 
-    // in all cases: avoid that arbitrary plugin toolviews get focus, like terminal, bug 412227
-    // we need to delay this a bit due to lazy view creation (and lazy e.g. terminal widget creation)
-    QTimer::singleShot(0, centralWidget(), SLOT(setFocus()));
-
     // kwrite doesn't want sidebars, we have no plugins
     if (KateApp::isKWrite()) {
         setSidebarsVisibleInternal(false, true);
     }
+
+    // trigger proper focus restore
+    m_viewManager->triggerActiveViewFocus();
 }
 
 KateMainWindow::~KateMainWindow()
@@ -179,6 +196,18 @@ KateMainWindow::~KateMainWindow()
     // save other options ;=)
     saveOptions();
 
+    // close all documents not visible in other windows, we did ask for permission in queryClose
+    if (winClosesDocuments()) {
+        auto docs = KateApp::self()->documentManager()->documentList();
+        docs.erase(std::remove_if(docs.begin(),
+                                  docs.end(),
+                                  [this](auto doc) {
+                                      return KateApp::self()->documentVisibleInOtherWindows(doc, this);
+                                  }),
+                   docs.end());
+        KateApp::self()->documentManager()->closeDocuments(docs, false);
+    }
+
     // unregister mainwindow in app
     KateApp::self()->removeMainWindow(this);
 
@@ -248,24 +277,18 @@ void KateMainWindow::setupImportantActions()
     actionCollection()->setDefaultShortcut(a, QKeySequence(Qt::CTRL | Qt::ALT | Qt::Key_O));
     connect(a, &QAction::triggered, this, &KateMainWindow::slotQuickOpen);
     a->setWhatsThis(i18n("Open a form to quick open documents."));
+
+    // enable hamburger menu
+    auto hamburgerMenu = static_cast<KHamburgerMenu *>(actionCollection()->addAction(KStandardAction::HamburgerMenu, QStringLiteral("hamburger_menu")));
+    hamburgerMenu->setMenuBar(menuBar());
+    hamburgerMenu->setShowMenuBarAction(m_paShowMenuBar);
 }
 
 void KateMainWindow::setupMainWindow()
 {
-    setToolViewStyle(KMultiTabBar::KDEV3ICON);
-
-    /**
-     * create central stacked widget with its children
-     */
-    m_mainStackedWidget = new QStackedWidget(centralWidget());
-    centralWidget()->layout()->addWidget(m_mainStackedWidget);
-    (static_cast<QBoxLayout *>(centralWidget()->layout()))->setStretchFactor(m_mainStackedWidget, 100);
-
-    m_viewManager = new KateViewManager(m_mainStackedWidget, this);
-    m_mainStackedWidget->addWidget(m_viewManager);
-
-    // make view manager default visible!
-    m_mainStackedWidget->setCurrentWidget(m_viewManager);
+    m_viewManager = new KateViewManager(centralWidget(), this);
+    centralWidget()->layout()->addWidget(m_viewManager);
+    (static_cast<QBoxLayout *>(centralWidget()->layout()))->setStretchFactor(m_viewManager, 100);
 
     m_bottomViewBarContainer = new QWidget(centralWidget());
     centralWidget()->layout()->addWidget(m_bottomViewBarContainer);
@@ -280,7 +303,7 @@ void KateMainWindow::setupMainWindow()
                                       KTextEditor::MainWindow::Bottom,
                                       QIcon::fromTheme(QStringLiteral("output_win")),
                                       i18n("Output"));
-    m_outputView = new KateOutputView(this, m_toolViewOutput);
+    m_outputView = new KateOutputView(this, m_toolViewOutput, toolviewToggleButton(static_cast<KateMDI::ToolView *>(m_toolViewOutput)));
 }
 
 void KateMainWindow::setupActions()
@@ -294,7 +317,12 @@ void KateMainWindow::setupActions()
         ->addAction(KStandardAction::Open, QStringLiteral("file_open"), m_viewManager, SLOT(slotDocumentOpen()))
         ->setWhatsThis(i18n("Open an existing document for editing"));
 
-    m_fileOpenRecent = KStandardAction::openRecent(m_viewManager, SLOT(openUrl(QUrl)), this);
+    m_fileOpenRecent = KStandardAction::openRecent(
+        m_viewManager,
+        [this](const QUrl &url) {
+            viewManager()->openUrlOrProject(url);
+        },
+        this);
     m_fileOpenRecent->setMaxItems(KateConfigDialog::recentFilesMaxCount());
     actionCollection()->addAction(m_fileOpenRecent->objectName(), m_fileOpenRecent);
     m_fileOpenRecent->setWhatsThis(i18n("This lists files which you have opened recently, and allows you to easily open them again."));
@@ -389,9 +417,10 @@ void KateMainWindow::setupActions()
 
     a = actionCollection()->addAction(QStringLiteral("view_new_view"));
     a->setIcon(QIcon::fromTheme(QStringLiteral("window-new")));
-    a->setText(i18n("&New Synchronized Window"));
+    a->setText(i18n("&New Window"));
+    actionCollection()->setDefaultShortcut(a, QKeySequence(Qt::CTRL | Qt::SHIFT | Qt::Key_N));
     connect(a, &QAction::triggered, this, &KateMainWindow::newWindow);
-    a->setWhatsThis(i18n("Create a new Kate window with a synchronized document list."));
+    a->setWhatsThis(i18n("Create a new window."));
 
     m_showFullScreenAction = KStandardAction::fullScreen(nullptr, nullptr, this, this);
     actionCollection()->addAction(m_showFullScreenAction->objectName(), m_showFullScreenAction);
@@ -424,7 +453,6 @@ void KateMainWindow::setupActions()
     }
 
     connect(m_viewManager, &KateViewManager::viewChanged, this, &KateMainWindow::slotWindowActivated);
-    connect(m_viewManager, &KateViewManager::viewChanged, this, &KateMainWindow::slotUpdateOpenWith);
     connect(m_viewManager, &KateViewManager::viewChanged, this, &KateMainWindow::slotUpdateActionsNeedingUrl);
     connect(m_viewManager, &KateViewManager::viewChanged, this, &KateMainWindow::slotUpdateBottomViewBar);
 
@@ -484,6 +512,7 @@ void KateMainWindow::setupActions()
     a = actionCollection()->addAction(QStringLiteral("view_history_back"));
     a->setIcon(QIcon::fromTheme(QStringLiteral("arrow-left")));
     a->setText(i18n("Go to Previous Location"));
+    actionCollection()->setDefaultShortcut(a, QKeySequence(Qt::CTRL | Qt::Key_1));
     connect(a, &QAction::triggered, this, [this] {
         m_viewManager->activeViewSpace()->goBack();
     });
@@ -492,10 +521,22 @@ void KateMainWindow::setupActions()
     a = actionCollection()->addAction(QStringLiteral("view_history_forward"));
     a->setIcon(QIcon::fromTheme(QStringLiteral("arrow-right")));
     a->setText(i18n("Go to Next Location"));
+    actionCollection()->setDefaultShortcut(a, QKeySequence(Qt::CTRL | Qt::SHIFT | Qt::Key_1));
     connect(a, &QAction::triggered, this, [this] {
         m_viewManager->activeViewSpace()->goForward();
     });
     connect(this->m_viewManager, &KateViewManager::historyForwardEnabled, a, &QAction::setEnabled);
+
+    a = actionCollection()->addAction(QStringLiteral("git_show_file_history"));
+    a->setText(i18n("Show File Git History"));
+    connect(a, &QAction::triggered, this, [this] {
+        if (activeView()) {
+            auto url = activeView()->document()->url();
+            if (url.isValid() && url.isLocalFile()) {
+                FileHistory::showFileHistory(url.toLocalFile(), m_wrapper);
+            }
+        }
+    });
 }
 
 void KateMainWindow::slotDocumentCloseAll()
@@ -547,7 +588,7 @@ void KateMainWindow::slotDocumentCloseOther()
     slotDocumentCloseOther(m_viewManager->activeView()->document());
 }
 
-bool KateMainWindow::queryClose_internal(KTextEditor::Document *doc)
+bool KateMainWindow::queryClose_internal(KTextEditor::Document *doc, KateMainWindow *win)
 {
     const auto documentCount = KateApp::self()->documentManager()->documentList().size();
 
@@ -556,20 +597,31 @@ bool KateMainWindow::queryClose_internal(KTextEditor::Document *doc)
     }
 
     std::vector<KTextEditor::Document *> modifiedDocuments = KateApp::self()->documentManager()->modifiedDocumentList();
-    modifiedDocuments.erase(std::remove(modifiedDocuments.begin(), modifiedDocuments.end(), doc), modifiedDocuments.end());
 
     // filter out what the stashManager will itself stash
-    auto m = modifiedDocuments.begin();
-    while (m != modifiedDocuments.end()) {
-        if (KateApp::self()->stashManager()->willStashDoc(*m)) {
-            m = modifiedDocuments.erase(m);
-        } else {
-            ++m;
-        }
+    modifiedDocuments.erase(std::remove_if(modifiedDocuments.begin(),
+                                           modifiedDocuments.end(),
+                                           [](auto doc) {
+                                               return KateApp::self()->stashManager()->willStashDoc(doc);
+                                           }),
+                            modifiedDocuments.end());
+
+    // do we want to ignore some document?
+    if (doc) {
+        modifiedDocuments.erase(std::remove(modifiedDocuments.begin(), modifiedDocuments.end(), doc), modifiedDocuments.end());
+    }
+
+    // do we want to ignore all documents visible in other windows?
+    if (win) {
+        modifiedDocuments.erase(std::remove_if(modifiedDocuments.begin(),
+                                               modifiedDocuments.end(),
+                                               [win](auto doc) {
+                                                   return KateApp::self()->documentVisibleInOtherWindows(doc, win);
+                                               }),
+                                modifiedDocuments.end());
     }
 
     bool shutdown = modifiedDocuments.empty();
-
     if (!shutdown) {
         shutdown = KateSaveModifiedDialog::queryClose(this, modifiedDocuments);
     }
@@ -594,9 +646,9 @@ bool KateMainWindow::queryClose()
     }
 
     // normal closing of window
-    // allow to close all windows until the last without restrictions
+    // if we are not the last window, just close the documents we own
     if (KateApp::self()->mainWindowsCount() > 1) {
-        return true;
+        return winClosesDocuments() ? queryClose_internal(nullptr, this) : true;
     }
 
     // last one: check if we can close all documents, try run
@@ -715,7 +767,8 @@ void KateMainWindow::toggleShowMenuBar(bool showMessage)
             m_viewManager->activeView()->contextMenu()->removeAction(m_paShowMenuBar);
         }
     } else {
-        if (showMessage) {
+        // we have a hamburger button in the toolbar, we can avoid the message if that is still visible
+        if (showMessage && toolBar()->isHidden()) {
             const QString accel = m_paShowMenuBar->shortcut().toString();
             KMessageBox::information(this,
                                      i18n("This will hide the menu bar completely."
@@ -758,24 +811,10 @@ void KateMainWindow::slotWindowActivated()
         updateCaption(m_viewManager->activeView()->document());
     }
 
-    // show view manager in any case
-    if (m_mainStackedWidget->currentWidget() != m_viewManager) {
-        m_mainStackedWidget->setCurrentWidget(m_viewManager);
-    }
-
     // update proxy
     centralWidget()->setFocusProxy(m_viewManager->activeView());
 }
 
-void KateMainWindow::slotUpdateOpenWith()
-{
-    if (m_viewManager->activeView()) {
-        documentOpenWith->setEnabled(!m_viewManager->activeView()->document()->url().isEmpty());
-    } else {
-        documentOpenWith->setEnabled(false);
-    }
-}
-
 void KateMainWindow::slotUpdateActionsNeedingUrl()
 {
     auto &&view = viewManager()->activeView();
@@ -786,6 +825,7 @@ void KateMainWindow::slotUpdateActionsNeedingUrl()
     action("file_rename")->setEnabled(hasUrl);
     action("file_delete")->setEnabled(hasUrl);
     action("file_properties")->setEnabled(hasUrl);
+    documentOpenWith->setEnabled(hasUrl);
 }
 
 void KateMainWindow::dragEnterEvent(QDragEnterEvent *event)
@@ -833,14 +873,22 @@ void KateMainWindow::slotDropEvent(QDropEvent *event)
             KFileItem kitem(url);
             kitem.setDelayedMimeTypes(true);
             if (kitem.isDir()) {
+#if KWIDGETSADDONS_VERSION >= QT_VERSION_CHECK(5, 100, 0)
+                if (KMessageBox::questionTwoActions(this,
+#else
                 if (KMessageBox::questionYesNo(this,
-                                               i18n("You dropped the directory %1 into Kate. "
-                                                    "Do you want to open all files contained in it?",
-                                                    url.url()),
-                                               i18nc("@title:window", "Open Files Recursively"),
-                                               KGuiItem(i18nc("@action:button", "Open All Files"), QStringLiteral("document-open")),
-                                               KStandardGuiItem::cancel())
+#endif
+                                                    i18n("You dropped the directory %1 into Kate. "
+                                                         "Do you want to open all files contained in it?",
+                                                         url.url()),
+                                                    i18nc("@title:window", "Open Files Recursively"),
+                                                    KGuiItem(i18nc("@action:button", "Open All Files"), QStringLiteral("document-open")),
+                                                    KStandardGuiItem::cancel())
+#if KWIDGETSADDONS_VERSION >= QT_VERSION_CHECK(5, 100, 0)
+                    == KMessageBox::PrimaryAction) {
+#else
                     == KMessageBox::Yes) {
+#endif
                     KIO::ListJob *list_job = KIO::listRecursive(url, KIO::DefaultFlags, false);
                     connect(list_job, &KIO::ListJob::entries, this, &KateMainWindow::slotListRecursiveEntries);
                 }
@@ -905,24 +953,31 @@ void KateMainWindow::slotConfigure()
 
 bool KateMainWindow::showPluginConfigPage(KTextEditor::Plugin *configpageinterface, int id)
 {
-    if (!m_viewManager->activeView()) {
-        return false;
+    KateConfigDialog *dlg = KateConfigDialog::widget(this);
+    if (dlg->mainwWindow() != this) {
+        dlg->reject();
+        if (dlg->mainwWindow())
+            dlg->mainwWindow()->removeWidget(dlg);
+        delete dlg;
+        dlg = KateConfigDialog::widget(this);
     }
-
-    KateConfigDialog *dlg = new KateConfigDialog(this);
     if (configpageinterface) {
         dlg->showAppPluginPage(configpageinterface, id);
     }
 
-    if (dlg->exec() == QDialog::Accepted) {
+    activateWidget(dlg);
+
+    connect(dlg, &QDialog::rejected, this, [this, dlg]() {
+        removeWidget(dlg);
+    });
+
+    connect(dlg, &KateConfigDialog::saved, this, [this]() {
         m_fileOpenRecent->setMaxItems(KateConfigDialog::recentFilesMaxCount());
-    }
-
-    delete dlg;
-
-    m_viewManager->reactivateActiveView(); // gui (toolbars...) needs to be updated, because
-    // of possible changes that the configure dialog
-    // could have done on it, specially for plugins.
+        // gui (toolbars...) needs to be updated, because
+        // of possible changes that the configure dialog
+        // could have done on it, specially for plugins.
+        m_viewManager->replugActiveView();
+    });
 
     return true;
 }
@@ -978,7 +1033,11 @@ void KateMainWindow::slotOpenWithMenuAction(QAction *a)
     // If app is null, ApplicationLauncherJob will invoke the open-with dialog
     auto *job = new KIO::ApplicationLauncherJob(app);
     job->setUrls(list);
+#if KIO_VERSION >= QT_VERSION_CHECK(5, 98, 0)
+    job->setUiDelegate(KIO::createDefaultJobUiDelegate(KJobUiDelegate::AutoHandlingEnabled, this));
+#else
     job->setUiDelegate(new KIO::JobUiDelegate(KJobUiDelegate::AutoHandlingEnabled, this));
+#endif
     job->start();
 }
 
@@ -1035,7 +1094,6 @@ void KateMainWindow::slotDocumentCreated(KTextEditor::Document *doc)
     connect(doc, &KTextEditor::Document::readWriteChanged, this, QOverload<KTextEditor::Document *>::of(&KateMainWindow::updateCaption));
     connect(doc, &KTextEditor::Document::documentNameChanged, this, QOverload<KTextEditor::Document *>::of(&KateMainWindow::updateCaption));
     connect(doc, &KTextEditor::Document::documentUrlChanged, this, QOverload<KTextEditor::Document *>::of(&KateMainWindow::updateCaption));
-    connect(doc, &KTextEditor::Document::documentUrlChanged, this, &KateMainWindow::slotUpdateOpenWith);
     connect(doc, &KTextEditor::Document::documentUrlChanged, this, &KateMainWindow::slotUpdateActionsNeedingUrl);
 
     updateCaption(doc);
@@ -1062,18 +1120,25 @@ void KateMainWindow::updateCaption(KTextEditor::Document *doc)
     }
 
     QString c;
+    const auto url = m_viewManager->activeView()->document()->url();
     if (m_viewManager->activeView()->document()->url().isEmpty() || (!m_paShowPath || !m_paShowPath->isChecked())) {
         c = m_viewManager->activeView()->document()->documentName();
     } else {
-        c = m_viewManager->activeView()->document()->url().toString(QUrl::PreferLocalFile);
-
-        const QString homePath = QDir::homePath();
-        if (c.startsWith(homePath)) {
-            c = QLatin1String("~") + c.right(c.length() - homePath.length());
+        // we want some filename @ folder output to have chance to keep important stuff even on elide
+        if (url.isLocalFile()) {
+            // perhaps shorten the path
+            const QString homePath = QDir::homePath();
+            QString path = url.toString(QUrl::RemoveFilename | QUrl::PreferLocalFile | QUrl::StripTrailingSlash);
+            if (path.startsWith(homePath)) {
+                path = QLatin1String("~") + path.right(path.length() - homePath.length());
+            }
+            c = url.fileName() + QStringLiteral(" @ ") + path;
+        } else {
+            c = url.toDisplayString();
         }
     }
 
-    setWindowFilePath(m_viewManager->activeView()->document()->url().toString(QUrl::PreferLocalFile));
+    setWindowFilePath(url.toString(QUrl::PreferLocalFile));
 
     QString sessName = KateApp::self()->sessionManager()->activeSession()->name();
     if (!sessName.isEmpty()) {
@@ -1293,10 +1358,69 @@ QObject *KateMainWindow::pluginView(const QString &name)
     return m_pluginViews.contains(plugin) ? m_pluginViews.value(plugin) : nullptr;
 }
 
-void KateMainWindow::addWidgetAsTab(QWidget *widget)
+bool KateMainWindow::addWidget(QWidget *widget)
 {
+    if (!widget) {
+        qWarning() << Q_FUNC_INFO << "Unexpected null widget!";
+        return false;
+    }
+
     auto vs = m_viewManager->activeViewSpace();
     vs->addWidgetAsTab(widget);
+    Q_EMIT widgetAdded(widget);
+    m_viewManager->activateView(widget);
+    return true;
+}
+
+bool KateMainWindow::removeWidget(QWidget *widget)
+{
+    return m_viewManager->removeWidget(widget);
+}
+
+QWidget *KateMainWindow::activeWidget()
+{
+    auto vs = m_viewManager->activeViewSpace();
+    if (auto w = vs->currentWidget()) {
+        return w;
+    }
+    return activeView();
+}
+
+void KateMainWindow::activateWidget(QWidget *widget)
+{
+    if (!m_viewManager->activateWidget(widget)) {
+        addWidget(widget);
+    }
+}
+
+void KateMainWindow::showDiff(const QByteArray &wordDiff, const DiffParams &params)
+{
+    auto w = new DiffWidget(params, this);
+    if (!params.tabTitle.isEmpty()) {
+        w->setWindowTitle(params.tabTitle);
+    } else {
+        if (params.destFile.isEmpty())
+            w->setWindowTitle(i18n("Diff %1", Utils::fileNameFromPath(params.srcFile)));
+        else
+            w->setWindowTitle(i18n("Diff %1..%2", Utils::fileNameFromPath(params.srcFile), Utils::fileNameFromPath(params.destFile)));
+    }
+
+    w->setWindowIcon(QIcon::fromTheme(QStringLiteral("text-x-patch")));
+    addWidget(w);
+    w->openDiff(wordDiff);
+}
+
+void KateMainWindow::showMessage(const QVariantMap &map)
+{
+    if (!m_outputView) {
+        return;
+    }
+    m_outputView->slotMessage(map);
+}
+
+QWidgetList KateMainWindow::widgets() const
+{
+    return m_viewManager->widgets();
 }
 
 void KateMainWindow::mousePressEvent(QMouseEvent *e)
@@ -1416,6 +1540,8 @@ void KateMainWindow::addRecentOpenedFile(const QUrl &url)
     // to our local list, aka menu
     m_fileOpenRecent->addUrl(url);
 
+    /** FIXME Disabled because this can be too slow 100ms/doc sometimes,
+     renable when it is 0/ms again*/
     // to the global "Recent Document Menu", see bug 420504
-    KRecentDocument::add(url);
+    // KRecentDocument::add(url);
 }
