@@ -13,11 +13,7 @@
 #include <QApplication>
 #include <QDataStream>
 #include <QDrag>
-#include <QIcon>
-#include <QMimeData>
-#include <QPainter>
 #include <QPixmap>
-#include <QResizeEvent>
 #include <QStyleOptionTab>
 #include <QStylePainter>
 #include <QWheelEvent>
@@ -28,12 +24,6 @@
 
 #include <KTextEditor/Document>
 
-struct KateTabButtonData {
-    KTextEditor::Document *doc = nullptr;
-};
-
-Q_DECLARE_METATYPE(KateTabButtonData)
-
 /**
  * Creates a new tab bar with the given \a parent.
  */
@@ -67,30 +57,49 @@ void KateTabBar::readConfig()
     KConfigGroup cgGeneral = KConfigGroup(config, "General");
 
     // 0 == unlimited, normalized other inputs
-    const int tabCountLimit = cgGeneral.readEntry("Tabbar Tab Limit", 0);
-    m_tabCountLimit = (tabCountLimit <= 0) ? 0 : tabCountLimit;
+    const int tabCountLimit = std::max(cgGeneral.readEntry("Tabbar Tab Limit", 0), 0);
+    if (m_tabCountLimit != tabCountLimit) {
+        m_tabCountLimit = tabCountLimit;
+        const QVector<DocOrWidget> docList = documentList();
+        if (m_tabCountLimit > 0 && docList.count() > m_tabCountLimit) {
+            // close N least used tabs
+            QMap<quint64, DocOrWidget> lruDocs;
+            for (const DocOrWidget &doc : docList) {
+                lruDocs[m_docToLruCounterAndHasTab[doc].first] = doc;
+            }
+            int toRemove = docList.count() - m_tabCountLimit;
+            for (DocOrWidget doc : std::as_const(lruDocs)) {
+                if (toRemove-- == 0) {
+                    break;
+                }
+                removeTab(documentIdx(doc));
+            }
+        } else if (m_docToLruCounterAndHasTab.size() > (size_t)docList.size()) {
+            // populate N recently user documents
+            std::map<quint64, DocOrWidget, std::greater<quint64>> mruDocs;
+            for (const auto &i : m_docToLruCounterAndHasTab) {
+                DocOrWidget doc = i.first;
+                if (!docList.contains(doc)) {
+                    mruDocs[i.second.first] = doc;
+                }
+            }
+            int toAdd = m_tabCountLimit - docList.count();
+            for (const auto &i : mruDocs) {
+                if (toAdd-- == 0) {
+                    break;
+                }
+                DocOrWidget doc = i.second;
+                const auto idx = addTab(doc.doc() ? doc.doc()->documentName() : doc.widget()->windowTitle());
+                setTabDocument(idx, doc);
+            }
+        }
+    }
 
     // use scroll buttons if we have no limit
     setUsesScrollButtons(m_tabCountLimit == 0 || cgGeneral.readEntry("Allow Tab Scrolling", true));
 
-    // elide if we have some limit
-    setElideMode((m_tabCountLimit == 0 || !cgGeneral.readEntry("Elide Tab Text", false)) ? Qt::ElideNone : Qt::ElideMiddle);
-
-    const std::vector<int> documentTabIndexes = this->documentTabIndexes();
-
-    // if we enforce a limit: purge tabs that violate it
-    if (m_tabCountLimit > 0 && ((int)documentTabIndexes.size() > m_tabCountLimit)) {
-        // just purge last X tabs, this isn't that clever but happens only on config changes!
-        size_t toRemove = documentTabIndexes.size() - (size_t)m_tabCountLimit;
-        size_t removed = 0;
-        for (auto rit = documentTabIndexes.rbegin(); rit != documentTabIndexes.rend(); ++rit) {
-            if (removed == toRemove)
-                break;
-            removeTab(*rit);
-            removed++;
-        }
-        setCurrentIndex(0);
-    }
+    // elide if requested, this is independent of the limit, just honor the users wish
+    setElideMode(cgGeneral.readEntry("Elide Tab Text", false) ? Qt::ElideMiddle : Qt::ElideNone);
 
     // handle tab close button and expansion
     setExpanding(cgGeneral.readEntry("Expand Tabs", false));
@@ -101,18 +110,6 @@ void KateTabBar::readConfig()
     m_middleClickCloseDocument = cgGeneral.readEntry("Tab Middle Click Close Document", true);
 }
 
-std::vector<int> KateTabBar::documentTabIndexes() const
-{
-    std::vector<int> docs;
-    const int tabCount = count();
-    for (int i = 0; i < tabCount; ++i) {
-        if (tabData(i).value<KateTabButtonData>().doc) {
-            docs.push_back(i);
-        }
-    }
-    return docs;
-}
-
 void KateTabBar::setActive(bool active)
 {
     if (active == m_isActive) {
@@ -147,7 +144,8 @@ bool KateTabBar::containsTab(int index) const
 QVariant KateTabBar::ensureValidTabData(int idx)
 {
     if (!tabData(idx).isValid()) {
-        setTabData(idx, QVariant::fromValue(KateTabButtonData{}));
+        DocOrWidget v(static_cast<KTextEditor::Document *>(nullptr));
+        setTabData(idx, QVariant::fromValue(v));
     }
     return tabData(idx);
 }
@@ -167,8 +165,11 @@ void KateTabBar::mousePressEvent(QMouseEvent *event)
         Q_EMIT activateViewSpaceRequested();
     }
 
-    if (event->button() == Qt::LeftButton && tabAt(event->pos()) != -1) {
+    int tab = tabAt(event->pos());
+    if (event->button() == Qt::LeftButton && tab != -1 && count() > 1) {
         dragStartPos = event->pos();
+        auto r = tabRect(tab);
+        dragHotspotPos = {dragStartPos.x() - r.x(), dragStartPos.y() - r.y()};
     } else {
         dragStartPos = {};
     }
@@ -197,7 +198,11 @@ void KateTabBar::mouseMoveEvent(QMouseEvent *event)
     // might not know that kate's tabs can be dragged to other
     // places, unless they drag it really far away
     auto viewSpace = qobject_cast<KateViewSpace *>(parentWidget());
-    if (!viewSpace || viewSpace->rect().contains(event->pos())) {
+    const auto viewspaceRect = viewSpace->rect();
+    QRect viewspaceRectTopArea = viewspaceRect;
+    viewspaceRectTopArea.setTop(viewspaceRect.top() - 40);
+    viewspaceRectTopArea.setBottom(viewspaceRect.height() / 2);
+    if (!viewSpace || viewspaceRectTopArea.contains(event->pos())) {
         QTabBar::mouseMoveEvent(event);
         return;
     }
@@ -211,10 +216,17 @@ void KateTabBar::mouseMoveEvent(QMouseEvent *event)
         return QTabBar::mouseMoveEvent(event);
     }
 
-    QDrag *drag = new QDrag(this);
+    int tab = currentIndex();
+    if (tab < 0) {
+        return;
+    }
 
-    int tab = tabAt(dragStartPos);
     QRect rect = tabRect(tab);
+    const auto tabData = this->tabData(tab).value<DocOrWidget>();
+    auto *tabObject = tabData.qobject();
+    if (!tabObject) {
+        return;
+    }
 
     QPixmap p(rect.size() * this->devicePixelRatioF());
     p.setDevicePixelRatio(this->devicePixelRatioF());
@@ -240,34 +252,60 @@ void KateTabBar::mouseMoveEvent(QMouseEvent *event)
     paint.drawControl(QStyle::CE_TabBarTab, opt);
     paint.end();
 
+    QByteArray data;
     auto view = viewSpace->currentView();
-    if (!view) {
+    if (view) {
+        KTextEditor::Cursor cp = view->cursorPosition();
+        QDataStream ds(&data, QIODevice::WriteOnly);
+        ds << cp.line();
+        ds << cp.column();
+        ds << view->document()->url();
+    } else if (!viewSpace->currentWidget()) {
+        qWarning() << "No view or widget, why?";
         return;
     }
 
-    QByteArray data;
-    KTextEditor::Cursor cp = view->cursorPosition();
-    QDataStream ds(&data, QIODevice::WriteOnly);
-    ds << cp.line();
-    ds << cp.column();
-    ds << view->document()->url();
-
-    auto mime = new TabMimeData(viewSpace, tabDocument(tab));
+    auto mime = new TabMimeData(viewSpace, tabData);
     mime->setData(QStringLiteral("application/kate.tab.mimedata"), data);
 
+    QDrag *drag = new QDrag(this);
     drag->setMimeData(mime);
     drag->setPixmap(p);
-    QPoint hp;
-    hp.setX(dragStartPos.x() - rect.x());
-    hp.setY(dragStartPos.y() - rect.y());
-    drag->setHotSpot(hp);
+    drag->setHotSpot(dragHotspotPos);
 
+    auto posCopy = dragStartPos;
     dragStartPos = {};
-    drag->exec();
+    dragHotspotPos = {};
+    drag->exec(Qt::CopyAction);
 
-    // We send this even to ensure the "moveable tab" is properly reset and we have no dislocated tabs
-    QMouseEvent *e = new QMouseEvent(QEvent::MouseButtonPress, QPoint(), Qt::LeftButton, Qt::LeftButton, Qt::NoModifier);
-    qApp->postEvent(this, e);
+    // On drag end we check whether the drag was a success
+    // i.e., if the widget/doc in tabData still exists in
+    // this tabbar, it failed or there was a copy operation
+    // in which case there might be the "movable tab" hanging
+    // in between which we reset
+    auto onDragEnd = [this, tabObject, posCopy]() {
+        bool found = false;
+        int tabIdx = 0;
+        for (; tabIdx < count(); ++tabIdx) {
+            auto d = this->tabData(tabIdx);
+            // We only expect doc, no dnd support for widgets
+            if (d.value<DocOrWidget>().qobject() == tabObject) {
+                found = true;
+                break;
+            }
+        }
+
+        if (found) {
+            // We send this even to ensure the "moveable tab" is properly reset and we have no dislocated tabs
+            auto e = QMouseEvent(QEvent::MouseButtonPress, posCopy, Qt::LeftButton, Qt::LeftButton, Qt::NoModifier);
+            qApp->sendEvent(this, &e);
+            if (currentIndex() != tabIdx) {
+                setCurrentIndex(tabIdx);
+            }
+        }
+    };
+
+    connect(drag, &QDrag::destroyed, this, onDragEnd);
 }
 
 void KateTabBar::contextMenuEvent(QContextMenuEvent *ev)
@@ -278,18 +316,21 @@ void KateTabBar::contextMenuEvent(QContextMenuEvent *ev)
     }
 }
 
-void KateTabBar::setTabDocument(int idx, KTextEditor::Document *doc)
+void KateTabBar::setTabDocument(int idx, DocOrWidget d)
 {
-    QVariant data = ensureValidTabData(idx);
-    KateTabButtonData buttonData = data.value<KateTabButtonData>();
-    buttonData.doc = doc;
-    setTabData(idx, QVariant::fromValue(buttonData));
+    setTabData(idx, QVariant::fromValue(d));
+
+    auto *doc = d.doc();
     // BUG: 441340 We need to escape the & because it is used for accelerators/shortcut mnemonic by default
-    QString tabName = doc->documentName();
+    QString tabName = d.doc() ? d.doc()->documentName() : d.widget()->windowTitle();
     tabName.replace(QLatin1Char('&'), QLatin1String("&&"));
     setTabText(idx, tabName);
-    setTabToolTip(idx, doc->url().toDisplayString());
-    setModifiedStateIcon(idx, doc);
+    if (d.doc()) {
+        setTabToolTip(idx, doc->url().isValid() ? doc->url().toDisplayString(QUrl::PreferLocalFile) : doc->documentName());
+        setModifiedStateIcon(idx, doc);
+    } else {
+        setTabIcon(idx, d.widget()->windowIcon());
+    }
 }
 
 void KateTabBar::setModifiedStateIcon(int idx, KTextEditor::Document *doc)
@@ -298,15 +339,15 @@ void KateTabBar::setModifiedStateIcon(int idx, KTextEditor::Document *doc)
     setTabIcon(idx, Utils::iconForDocument(doc));
 }
 
-void KateTabBar::setCurrentDocument(KTextEditor::Document *doc)
+void KateTabBar::setCurrentDocument(DocOrWidget docOrWidget)
 {
     // in any case: update lru counter for this document, might add new element to hash
     // we have a tab after this call, too!
-    m_docToLruCounterAndHasTab[doc] = std::make_pair(++m_lruCounter, true);
+    m_docToLruCounterAndHasTab[docOrWidget] = std::make_pair(++m_lruCounter, true);
 
     // do we have a tab for this document?
     // if yes => just set as current one
-    const int existingIndex = documentIdx(doc);
+    const int existingIndex = documentIdx(docOrWidget);
     if (existingIndex != -1) {
         setCurrentIndex(existingIndex);
         return;
@@ -314,9 +355,9 @@ void KateTabBar::setCurrentDocument(KTextEditor::Document *doc)
 
     // else: if we are still inside the allowed number of tabs or have no limit
     // => create new tab and be done
-    if ((m_tabCountLimit == 0) || documentTabIndexes().size() < (size_t)m_tabCountLimit) {
-        m_beingAdded = doc;
-        insertTab(-1, doc->documentName());
+    if ((m_tabCountLimit == 0) || (count() < m_tabCountLimit)) {
+        m_beingAdded = docOrWidget;
+        insertTab(-1, docOrWidget.doc() ? docOrWidget.doc()->documentName() : docOrWidget.widget()->windowTitle());
         return;
     }
 
@@ -326,13 +367,10 @@ void KateTabBar::setCurrentDocument(KTextEditor::Document *doc)
     // search for the right tab
     quint64 minCounter = static_cast<quint64>(-1);
     int indexToReplace = 0;
-    KTextEditor::Document *docToReplace = nullptr;
+    DocOrWidget docToReplace = DocOrWidget::null();
     for (int idx = 0; idx < count(); idx++) {
         QVariant data = tabData(idx);
-        KTextEditor::Document *doc = data.value<KateTabButtonData>().doc;
-        if (!data.isValid() || !doc) {
-            continue;
-        }
+        DocOrWidget doc = data.value<DocOrWidget>();
         const quint64 currentCounter = m_docToLruCounterAndHasTab[doc].first;
         if (currentCounter <= minCounter) {
             minCounter = currentCounter;
@@ -345,11 +383,11 @@ void KateTabBar::setCurrentDocument(KTextEditor::Document *doc)
     m_docToLruCounterAndHasTab[docToReplace].second = false;
 
     // replace it's data + set it as active
-    setTabDocument(indexToReplace, doc);
+    setTabDocument(indexToReplace, docOrWidget);
     setCurrentIndex(indexToReplace);
 }
 
-void KateTabBar::removeDocument(KTextEditor::Document *doc)
+void KateTabBar::removeDocument(DocOrWidget doc)
 {
     // purge LRU storage, must work
     auto erased = (m_docToLruCounterAndHasTab.erase(doc) == 1);
@@ -366,7 +404,7 @@ void KateTabBar::removeDocument(KTextEditor::Document *doc)
     // if we have some tab limit, replace the removed tab with the next best document that has none!
     if (m_tabCountLimit > 0) {
         quint64 maxCounter = 0;
-        KTextEditor::Document *docToReplace = nullptr;
+        DocOrWidget docToReplace = DocOrWidget::null();
         for (const auto &lru : m_docToLruCounterAndHasTab) {
             // ignore stuff with tabs
             if (lru.second.second) {
@@ -381,7 +419,7 @@ void KateTabBar::removeDocument(KTextEditor::Document *doc)
         }
 
         // any document found? replace the tab we want to close and be done
-        if (docToReplace) {
+        if (!docToReplace.isNull()) {
             // mark the replace doc as "has a tab"
             m_docToLruCounterAndHasTab[docToReplace].second = true;
 
@@ -398,35 +436,31 @@ void KateTabBar::removeDocument(KTextEditor::Document *doc)
     removeTab(idx);
 }
 
-int KateTabBar::documentIdx(KTextEditor::Document *doc)
+int KateTabBar::documentIdx(DocOrWidget doc)
 {
     for (int idx = 0; idx < count(); idx++) {
-        QVariant data = tabData(idx);
-        if (!data.isValid()) {
-            continue;
+        const QVariant data = tabData(idx);
+        if (data.value<DocOrWidget>().qobject() == doc.qobject()) {
+            return idx;
         }
-        if (data.value<KateTabButtonData>().doc != doc) {
-            continue;
-        }
-        return idx;
     }
     return -1;
 }
 
-KTextEditor::Document *KateTabBar::tabDocument(int idx)
+DocOrWidget KateTabBar::tabDocument(int idx)
 {
     QVariant data = ensureValidTabData(idx);
-    KateTabButtonData buttonData = data.value<KateTabButtonData>();
+    DocOrWidget buttonData = data.value<DocOrWidget>();
 
-    KTextEditor::Document *doc = nullptr;
-    // The tab got activated before the correct finalixation,
+    auto doc = DocOrWidget::null();
+    // The tab got activated before the correct finalization,
     // we need to plug the document before returning.
-    if (buttonData.doc == nullptr && m_beingAdded) {
+    if (buttonData.isNull() && !m_beingAdded.isNull()) {
         setTabDocument(idx, m_beingAdded);
         doc = m_beingAdded;
-        m_beingAdded = nullptr;
+        m_beingAdded.clear();
     } else {
-        doc = buttonData.doc;
+        doc = buttonData;
     }
 
     return doc;
@@ -434,29 +468,19 @@ KTextEditor::Document *KateTabBar::tabDocument(int idx)
 
 void KateTabBar::tabInserted(int idx)
 {
-    if (m_beingAdded) {
+    if (m_beingAdded.qobject()) {
         setTabDocument(idx, m_beingAdded);
     }
-    m_beingAdded = nullptr;
+    m_beingAdded.clear();
 }
 
-QVector<KTextEditor::Document *> KateTabBar::documentList() const
+QVector<DocOrWidget> KateTabBar::documentList() const
 {
-    QVector<KTextEditor::Document *> result;
+    QVector<DocOrWidget> result;
+    result.reserve(count());
     for (int idx = 0; idx < count(); idx++) {
         QVariant data = tabData(idx);
-        if (!data.isValid()) {
-            continue;
-        }
-        if (data.value<KateTabButtonData>().doc)
-            result.append(data.value<KateTabButtonData>().doc);
+        result.append(data.value<DocOrWidget>());
     }
     return result;
 }
-
-void KateTabBar::setCurrentWidget(QWidget *widget)
-{
-    int idx = insertTab(-1, widget->windowTitle());
-    setTabData(idx, QVariant::fromValue(widget));
-    setCurrentIndex(idx);
-}
